from dataclasses import dataclass
from enum import Enum
from typing import Optional

from randovania.bitpacking.json_dataclass import JsonDataclass
from randovania.game_description.area_location import AreaLocation
from randovania.game_description.resources.pickup_index import PickupIndex


class HintType(Enum):
    # Joke
    JOKE = "joke"

    # Hints for where a set of red temple keys are
    RED_TEMPLE_KEY_SET = "red-temple-key-set"

    # All other hints
    LOCATION = "location"


class HintDarkTemple(Enum):
    AGON_WASTES = "agon-wastes"
    TORVUS_BOG = "torvus-bog"
    SANCTUARY_FORTRESS = "sanctuary-fortress"


class HintItemPrecision(Enum):
    # The exact item
    DETAILED = "detailed"

    # movement, morph ball, etc
    PRECISE_CATEGORY = "precise-category"

    # major item, key, expansion
    GENERAL_CATEGORY = "general-category"

    # x-related, life-support, or just the precise category
    BROAD_CATEGORY = "broad-category"

    # Say nothing at all about the item
    NOTHING = "nothing"


class HintLocationPrecision(Enum):
    # The exact location
    DETAILED = "detailed"

    # Includes only the world of the location
    WORLD_ONLY = "world-only"

    # Keybearer corpses
    KEYBEARER = "keybearer"

    # Amorbis, Chykka, and Quadraxis
    GUARDIAN = "guardian"

    # Vanilla Light Suit location
    LIGHT_SUIT_LOCATION = "light-suit-location"

    RELATIVE_TO_AREA = "relative-to-area"
    RELATIVE_TO_INDEX = "relative-to-index"


class HintRelativeAreaName(Enum):
    # The area's name
    NAME = "name"

    # Some unique feature of the area
    FEATURE = "feature"


@dataclass(frozen=True)
class RelativeData:
    precise_distance: bool

    @classmethod
    def from_json(cls, param: dict) -> "RelativeData":
        if "area_location" in param:
            return RelativeDataArea.from_json(param)
        else:
            return RelativeDataItem.from_json(param)


@dataclass(frozen=True)
class RelativeDataItem(JsonDataclass, RelativeData):
    other_index: PickupIndex
    precision: HintItemPrecision


@dataclass(frozen=True)
class RelativeDataArea(JsonDataclass, RelativeData):
    area_location: AreaLocation
    precision: HintRelativeAreaName


@dataclass(frozen=True)
class PrecisionPair(JsonDataclass):
    location: HintLocationPrecision
    item: HintItemPrecision
    include_owner: bool
    relative: Optional[RelativeData] = None


@dataclass(frozen=True)
class Hint(JsonDataclass):
    hint_type: HintType
    precision: Optional[PrecisionPair]
    target: Optional[PickupIndex] = None
    dark_temple: Optional[HintDarkTemple] = None

    def __post_init__(self):
        if self.hint_type is HintType.JOKE:
            if self.target is not None or self.dark_temple is not None:
                raise ValueError(f"Joke Hint, but had a target or dark_temple.")
        elif self.hint_type is HintType.LOCATION:
            if self.target is None:
                raise ValueError(f"Location Hint, but no target set.")
        elif self.hint_type is HintType.RED_TEMPLE_KEY_SET:
            if self.dark_temple is None:
                raise ValueError(f"Dark Temple Hint, but no dark_temple set.")



import datetime
from typing import List

from PySide2.QtCore import Qt
from PySide2.QtWidgets import QPushButton, QDialogButtonBox, QDialog, QTableWidgetItem, QInputDialog, QLineEdit
from asyncqt import asyncSlot

from randovania.gui.generated.game_session_browser_dialog_ui import Ui_GameSessionBrowserDialog
from randovania.gui.lib import common_qt_lib, async_dialog
from randovania.gui.lib.qt_network_client import handle_network_errors, QtNetworkClient
from randovania.network_client.game_session import GameSessionListEntry
from randovania.network_client.network_client import ConnectionState
from randovania.network_common.error import WrongPassword
from randovania.network_common.session_state import GameSessionState


class GameSessionBrowserDialog(QDialog, Ui_GameSessionBrowserDialog):
    sessions: List[GameSessionListEntry]
    visible_sessions: List[GameSessionListEntry]

    def __init__(self, network_client: QtNetworkClient):
        super().__init__()
        self.setupUi(self)
        common_qt_lib.set_default_window_icon(self)
        self.network_client = network_client

        self.refresh_button = QPushButton("Refresh")
        self.button_box.addButton(self.refresh_button, QDialogButtonBox.ResetRole)
        self.button_box.button(QDialogButtonBox.Ok).setEnabled(False)
        self.button_box.button(QDialogButtonBox.Ok).setText("Join")

        self.button_box.accepted.connect(self.attempt_join)
        self.button_box.rejected.connect(self.reject)
        self.refresh_button.clicked.connect(self.refresh)

        checks = (
            self.has_password_yes_check,
            self.has_password_no_check,
            self.state_setup_check,
            self.state_inprogress_check,
            self.state_finished_check,
            self.filter_age_check,
        )
        for check in checks:
            check.stateChanged.connect(self.update_list)
        self.filter_name_edit.textEdited.connect(self.update_list)
        self.filter_age_spin.valueChanged.connect(self.update_list)

        self.table_widget.itemSelectionChanged.connect(self.on_selection_changed)
        self.table_widget.itemDoubleClicked.connect(self.on_double_click)

        self.network_client.ConnectionStateUpdated.connect(self.on_server_connection_state_updated)
        self.on_server_connection_state_updated(self.network_client.connection_state)

    @asyncSlot()
    @handle_network_errors
    async def refresh(self):
        self.refresh_button.setEnabled(False)
        try:
            self.sessions = await self.network_client.get_game_session_list()
            self.update_list()
        finally:
            self.refresh_button.setEnabled(True)

    def on_selection_changed(self):
        self.button_box.button(QDialogButtonBox.Ok).setEnabled(len(self.table_widget.selectedItems()) > 0)

    @property
    def selected_session(self) -> GameSessionListEntry:
        return self.table_widget.selectedItems()[0].data(Qt.UserRole)

    @asyncSlot(QTableWidgetItem)
    async def on_double_click(self, item: QTableWidgetItem):
        await self.attempt_join()

    @asyncSlot()
    @handle_network_errors
    async def attempt_join(self):
        if not self.visible_sessions:
            return

        session = self.selected_session

        if session.has_password:
            dialog = QInputDialog(self)
            dialog.setWindowTitle("Enter password")
            dialog.setLabelText("This session requires a password:")
            dialog.setWindowModality(Qt.WindowModal)
            dialog.setTextEchoMode(QLineEdit.Password)

            if await async_dialog.execute_dialog(dialog) != dialog.Accepted:
                return

            password = dialog.textValue()
        else:
            password = None

        try:
            await self.network_client.join_game_session(session, password)
            return self.accept()

        except WrongPassword:
            await async_dialog.warning(self, "Incorrect Password", "The password entered was incorrect.")

    def update_list(self):
        self.table_widget.clear()
        self.table_widget.setHorizontalHeaderLabels(["Name", "State", "Players", "Password?", "Creator",
                                                     "Creation Date"])

        name_filter = self.filter_name_edit.text().strip()

        displayed_has_password = set()
        if self.has_password_yes_check.isChecked():
            displayed_has_password.add(True)
        if self.has_password_no_check.isChecked():
            displayed_has_password.add(False)

        displayed_states = set()
        for (check, state) in ((self.state_setup_check, GameSessionState.SETUP),
                               (self.state_inprogress_check, GameSessionState.IN_PROGRESS),
                               (self.state_finished_check, GameSessionState.FINISHED)):
            if check.isChecked():
                displayed_states.add(state)

        dont_filter_age = not self.filter_age_check.isChecked()
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        max_session_age = datetime.timedelta(days=self.filter_age_spin.value())

        visible_sessions = [
            session
            for session in reversed(self.sessions)
            if (session.has_password in displayed_has_password
                and session.state in displayed_states
                and name_filter in session.name
                and (dont_filter_age or (now - session.creation_date) < max_session_age))
        ]
        self.visible_sessions = visible_sessions

        self.table_widget.setRowCount(len(visible_sessions))
        for i, session in enumerate(visible_sessions):
            name = QTableWidgetItem(session.name)
            state = QTableWidgetItem(session.state.user_friendly_name)
            players_item = QTableWidgetItem(str(session.num_players))
            has_password = QTableWidgetItem("Yes" if session.has_password else "No")
            creator = QTableWidgetItem(session.creator)
            creation_date = QTableWidgetItem(session.creation_date.astimezone().strftime("%Y-%m-%d %H:%M"))

            name.setData(Qt.UserRole, session)
            self.table_widget.setItem(i, 0, name)
            self.table_widget.setItem(i, 1, state)
            self.table_widget.setItem(i, 2, players_item)
            self.table_widget.setItem(i, 3, has_password)
            self.table_widget.setItem(i, 4, creator)
            self.table_widget.setItem(i, 5, creation_date)

        for i in range(6):
            self.table_widget.resizeColumnToContents(i)

        self.status_label.setText(f"{len(self.sessions)} sessions total, {len(visible_sessions)} displayed.")

    def on_server_connection_state_updated(self, state: ConnectionState):
        self.server_connection_label.setText(f"Server: {state.value}")



import pytest

import randovania.games.prime.echoes_items
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.generator.item_pool import sky_temple_keys, pickup_creator
from randovania.layout.echoes_configuration import LayoutSkyTempleKeyMode


def test_sky_temple_key_distribution_logic_all_bosses_valid(echoes_resource_database):
    # Run
    results = sky_temple_keys.add_sky_temple_key_distribution_logic(echoes_resource_database,
                                                                    LayoutSkyTempleKeyMode.ALL_BOSSES)
    item_pool, pickup_assignment, initial_items = results

    # Assert
    assert item_pool == []
    assert initial_items == {}
    assert list(pickup_assignment.keys()) == sky_temple_keys._GUARDIAN_INDICES + sky_temple_keys._SUB_GUARDIAN_INDICES


def test_sky_temple_key_distribution_logic_all_guardians_valid(echoes_resource_database):
    # Run
    results = sky_temple_keys.add_sky_temple_key_distribution_logic(echoes_resource_database,
                                                                    LayoutSkyTempleKeyMode.ALL_GUARDIANS)
    item_pool, pickup_assignment, initial_items = results

    # Assert
    assert item_pool == []
    assert initial_items == {
        ItemResourceInfo(randovania.games.prime.echoes_items.SKY_TEMPLE_KEY_ITEMS[i - 1],
                         f'Sky Temple Key {i}', f'TempleKey{i}', 1, None): 1
        for i in range(4, 10)
    }
    assert list(pickup_assignment.keys()) == sky_temple_keys._GUARDIAN_INDICES


@pytest.mark.parametrize("quantity", list(range(10)))
def test_sky_temple_key_distribution_logic_with_quantity(echoes_resource_database, quantity: int):
    # Run
    results = sky_temple_keys.add_sky_temple_key_distribution_logic(echoes_resource_database,
                                                                    LayoutSkyTempleKeyMode(quantity))
    item_pool, pickup_assignment, initial_items = results

    # Assert
    assert item_pool == [
        pickup_creator.create_sky_temple_key(i, echoes_resource_database)
        for i in range(quantity)
    ]
    assert pickup_assignment == {}
    assert initial_items == {
        ItemResourceInfo(randovania.games.prime.echoes_items.SKY_TEMPLE_KEY_ITEMS[i - 1],
                         f'Sky Temple Key {i}', f'TempleKey{i}', 1, None): 1
        for i in range(quantity + 1, 10)
    }



from randovania.game_description.resources.resource_type import ResourceType


class LogbookAsset:
    _string_asset_id: int

    @property
    def resource_type(self) -> ResourceType:
        return ResourceType.LOGBOOK_INDEX

    @property
    def long_name(self) -> str:
        return "LogbookAsset 0x{:08X}".format(self._string_asset_id)

    def __init__(self, asset_id: int):
        self._string_asset_id = asset_id

    def __lt__(self, other: "LogbookAsset") -> bool:
        return self._string_asset_id < other._string_asset_id

    def __repr__(self):
        return self.long_name

    def __hash__(self):
        return self._string_asset_id

    def __eq__(self, other):
        return isinstance(other, LogbookAsset) and other._string_asset_id == self._string_asset_id

    @property
    def asset_id(self) -> int:
        return self._string_asset_id



import random
from random import Random
from typing import List
from unittest.mock import patch, MagicMock

import pytest

from randovania.game_description.area_location import AreaLocation
from randovania.generator import elevator_distributor
from randovania.generator.elevator_distributor import Elevator


@pytest.mark.parametrize(["seed_number", "expected_ids"], [
    (5000, [589949, 2949235, 4260106, 2162826, 393260, 4522032, 152, 3538975, 136970379, 204865660, 589851,
            1245332, 38, 1572998, 1638535, 2097251, 3342446, 1245307, 122, 129, 524321, 1966093]),
    (9000, [1245307, 2949235, 1245332, 152, 589949, 4522032, 129, 2097251, 204865660, 1638535, 136970379,
            4260106, 589851, 1572998, 3538975, 38, 3342446, 2162826, 1966093, 524321, 393260, 122]),
    (1157772449, [4260106, 393260, 1638535, 2162826, 1245307, 2949235, 152, 3538975, 1572998, 129, 1245332,
                  204865660, 136970379, 3342446, 524321, 2097251, 38, 4522032, 589949, 1966093, 122, 589851])
])
def test_try_randomize_elevators(seed_number: int,
                                 expected_ids: List[int],
                                 echoes_game_description):
    # Setup
    rng = Random(seed_number)

    # Run
    result = elevator_distributor.try_randomize_elevators(
        rng,
        elevator_distributor.create_elevator_database(echoes_game_description.world_list, set()))
    connected_ids = [elevator.connected_elevator.instance_id for elevator in result]

    # Assert
    assert connected_ids == expected_ids


@patch("randovania.generator.elevator_distributor.try_randomize_elevators", autospec=True)
def test_two_way_elevator_connections_between_areas(mock_try_randomize_elevators: MagicMock,
                                                    ):
    # Setup
    rng = MagicMock()
    elevator = MagicMock()
    database = MagicMock()
    mock_try_randomize_elevators.return_value = [
        elevator
    ]

    # Run
    result = elevator_distributor.two_way_elevator_connections(rng, database, True)

    # Assert
    mock_try_randomize_elevators.assert_called_once_with(rng, database)
    assert result == {
        elevator.instance_id: elevator.connected_elevator.area_location,
    }


def test_two_way_elevator_connections_unchecked():
    # Setup
    rng = random.Random(5000)
    elevators = [
        Elevator(i, i, i, i, i)
        for i in range(6)
    ]
    database = tuple(elevators)

    # Run
    result = elevator_distributor.two_way_elevator_connections(rng, database, False)

    # Assert
    assert result == {
        0: AreaLocation(4, 4),
        1: AreaLocation(2, 2),
        2: AreaLocation(1, 1),
        3: AreaLocation(5, 5),
        4: AreaLocation(0, 0),
        5: AreaLocation(3, 3),
    }


def test_one_way_elevator_connections_elevator_target():
    # Setup
    rng = random.Random(5000)
    elevators = [
        Elevator(i, i, i, i, i)
        for i in range(6)
    ]
    database = tuple(elevators)

    # Run
    result = elevator_distributor.one_way_elevator_connections(rng, database, None, True)

    # Assert
    assert result == {
        0: AreaLocation(1, 1),
        1: AreaLocation(2, 2),
        2: AreaLocation(3, 3),
        3: AreaLocation(5, 5),
        4: AreaLocation(0, 0),
        5: AreaLocation(4, 4),
    }


def test_one_way_elevator_connections_any_target(echoes_game_description):
    # Setup
    rng = random.Random(5000)
    elevators = [
        Elevator(i, i, i, i, i)
        for i in range(6)
    ]
    database = tuple(elevators)

    # Run
    result = elevator_distributor.one_way_elevator_connections(rng, database, echoes_game_description.world_list, False)

    # Assert
    assert result == {
        0: AreaLocation(0x42B935E4, 0xA4B2CB7E),
        1: AreaLocation(0x3DFD2249, 0xCB165BD8),
        2: AreaLocation(0x3DFD2249, 0x63190A61),
        3: AreaLocation(0x1BAA96C2, 0xA2406387),
        4: AreaLocation(0x3BFA3EFF, 0x531079BA),
        5: AreaLocation(0x1BAA96C2, 0x844A690C),
    }



from typing import Callable, Union
from unittest.mock import MagicMock, patch, call

import randovania
from randovania.generator import generator
from randovania.layout.layout_description import LayoutDescription


@patch("randovania.generator.generator._validate_item_pool_size", autospec=True)
@patch("randovania.generator.generator.create_player_pool", autospec=True)
@patch("randovania.generator.generator._distribute_remaining_items", autospec=True)
@patch("randovania.generator.generator.run_filler", autospec=True)
@patch("randovania.generator.generator.Random", autospec=False)
def test_create_patches(mock_random: MagicMock,
                        mock_run_filler: MagicMock,
                        mock_distribute_remaining_items: MagicMock,
                        mock_create_player_pool: MagicMock,
                        mock_validate_item_pool_size: MagicMock,
                        ):
    # Setup
    num_players = 1
    rng = mock_random.return_value
    status_update: Union[MagicMock, Callable[[str], None]] = MagicMock()
    player_pools = [MagicMock() for _ in range(num_players)]
    presets = [MagicMock() for _ in range(num_players)]

    mock_create_player_pool.side_effect = player_pools

    permalink = MagicMock()
    permalink.get_preset.side_effect = lambda i: presets[i]

    # Run
    result = generator._async_create_description(permalink, status_update, 0)

    # Assert
    mock_random.assert_called_once_with(permalink.as_bytes)
    mock_create_player_pool.assert_has_calls([
        call(rng, presets[i].configuration, i, num_players)
        for i in range(num_players)
    ])
    mock_validate_item_pool_size.assert_has_calls([
        call(player_pools[i].pickups, player_pools[i].game, player_pools[i].configuration)
        for i in range(num_players)
    ])
    mock_run_filler.assert_called_once_with(rng, {i: player_pools[i] for i in range(num_players)}, status_update)
    mock_distribute_remaining_items.assert_called_once_with(rng, mock_run_filler.return_value.player_results)

    assert result == LayoutDescription(
        permalink=permalink,
        version=randovania.VERSION,
        all_patches=mock_distribute_remaining_items.return_value,
        item_order=mock_run_filler.return_value.action_log,
    )



import argparse
import os
import sys
from pathlib import Path

import randovania
from randovania.cli import echoes, server, gui, prime_database


def create_subparsers(root_parser):
    echoes.create_subparsers(root_parser)
    prime_database.create_subparsers(root_parser)
    server.create_subparsers(root_parser)
    gui.create_subparsers(root_parser)


def _print_version(args):
    print("Randovania {} from {}".format(
        randovania.VERSION,
        os.path.dirname(randovania.__file__)))


def _create_parser():
    parser = argparse.ArgumentParser()

    create_subparsers(parser.add_subparsers(dest="game"))
    parser.add_argument("--version", action="store_const",
                        const=_print_version, dest="func")
    parser.add_argument("--configuration", type=Path,
                        help="Use the given configuration path instead of the included one.")

    return parser


def _run_args(parser, args):
    if args.configuration is not None:
        randovania.CONFIGURATION_FILE_PATH = args.configuration.absolute()

    if args.func is None:
        parser.print_help()
        raise SystemExit(1)

    args.func(args)


def run_pytest(argv):
    import pytest
    import pytest_asyncio.plugin
    import pytest_mock.plugin
    sys.exit(pytest.main(argv[2:], plugins=[pytest_asyncio.plugin, pytest_mock.plugin]))


def run_cli(argv):
    if len(argv) > 1 and argv[1] == "--pytest":
        run_pytest(argv)
    else:
        args = argv[1:]
        if gui.has_gui and not args:
            args = ["gui", "main"]

        parser = _create_parser()
        _run_args(parser, parser.parse_args(args))



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_echoes_patches.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_PresetEchoesPatches(object):
    def setupUi(self, PresetEchoesPatches):
        if not PresetEchoesPatches.objectName():
            PresetEchoesPatches.setObjectName(u"PresetEchoesPatches")
        PresetEchoesPatches.resize(466, 454)
        self.centralWidget = QWidget(PresetEchoesPatches)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(self.centralWidget)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.include_menu_mod_check = QCheckBox(self.centralWidget)
        self.include_menu_mod_check.setObjectName(u"include_menu_mod_check")
        self.include_menu_mod_check.setEnabled(True)

        self.verticalLayout.addWidget(self.include_menu_mod_check)

        self.include_menu_mod_label = QLabel(self.centralWidget)
        self.include_menu_mod_label.setObjectName(u"include_menu_mod_label")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.include_menu_mod_label.sizePolicy().hasHeightForWidth())
        self.include_menu_mod_label.setSizePolicy(sizePolicy)
        self.include_menu_mod_label.setMaximumSize(QSize(16777215, 60))
        self.include_menu_mod_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.include_menu_mod_label.setWordWrap(True)
        self.include_menu_mod_label.setOpenExternalLinks(True)

        self.verticalLayout.addWidget(self.include_menu_mod_label)

        self.line_1 = QFrame(self.centralWidget)
        self.line_1.setObjectName(u"line_1")
        self.line_1.setFrameShape(QFrame.HLine)
        self.line_1.setFrameShadow(QFrame.Sunken)

        self.verticalLayout.addWidget(self.line_1)

        self.warp_to_start_check = QCheckBox(self.centralWidget)
        self.warp_to_start_check.setObjectName(u"warp_to_start_check")

        self.verticalLayout.addWidget(self.warp_to_start_check)

        self.warp_to_start_label = QLabel(self.centralWidget)
        self.warp_to_start_label.setObjectName(u"warp_to_start_label")
        sizePolicy.setHeightForWidth(self.warp_to_start_label.sizePolicy().hasHeightForWidth())
        self.warp_to_start_label.setSizePolicy(sizePolicy)
        self.warp_to_start_label.setMaximumSize(QSize(16777215, 70))
        self.warp_to_start_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.warp_to_start_label.setWordWrap(True)

        self.verticalLayout.addWidget(self.warp_to_start_label)

        self.line_2 = QFrame(self.centralWidget)
        self.line_2.setObjectName(u"line_2")
        self.line_2.setFrameShape(QFrame.HLine)
        self.line_2.setFrameShadow(QFrame.Sunken)

        self.verticalLayout.addWidget(self.line_2)

        self.skip_final_bosses_check = QCheckBox(self.centralWidget)
        self.skip_final_bosses_check.setObjectName(u"skip_final_bosses_check")

        self.verticalLayout.addWidget(self.skip_final_bosses_check)

        self.skip_final_bosses_label = QLabel(self.centralWidget)
        self.skip_final_bosses_label.setObjectName(u"skip_final_bosses_label")
        self.skip_final_bosses_label.setMouseTracking(True)
        self.skip_final_bosses_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.skip_final_bosses_label.setWordWrap(True)

        self.verticalLayout.addWidget(self.skip_final_bosses_label)

        self.line_3 = QFrame(self.centralWidget)
        self.line_3.setObjectName(u"line_3")
        self.line_3.setFrameShape(QFrame.HLine)
        self.line_3.setFrameShadow(QFrame.Sunken)

        self.verticalLayout.addWidget(self.line_3)

        self.pickup_style_box = QGroupBox(self.centralWidget)
        self.pickup_style_box.setObjectName(u"pickup_style_box")
        sizePolicy.setHeightForWidth(self.pickup_style_box.sizePolicy().hasHeightForWidth())
        self.pickup_style_box.setSizePolicy(sizePolicy)
        self.verticalLayout_2 = QVBoxLayout(self.pickup_style_box)
        self.verticalLayout_2.setSpacing(6)
        self.verticalLayout_2.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.pickup_model_combo = QComboBox(self.pickup_style_box)
        self.pickup_model_combo.addItem("")
        self.pickup_model_combo.addItem("")
        self.pickup_model_combo.addItem("")
        self.pickup_model_combo.addItem("")
        self.pickup_model_combo.setObjectName(u"pickup_model_combo")

        self.verticalLayout_2.addWidget(self.pickup_model_combo)

        self.pickup_data_source_label = QLabel(self.pickup_style_box)
        self.pickup_data_source_label.setObjectName(u"pickup_data_source_label")

        self.verticalLayout_2.addWidget(self.pickup_data_source_label)

        self.pickup_data_source_combo = QComboBox(self.pickup_style_box)
        self.pickup_data_source_combo.addItem("")
        self.pickup_data_source_combo.addItem("")
        self.pickup_data_source_combo.addItem("")
        self.pickup_data_source_combo.setObjectName(u"pickup_data_source_combo")

        self.verticalLayout_2.addWidget(self.pickup_data_source_combo)


        self.verticalLayout.addWidget(self.pickup_style_box)

        PresetEchoesPatches.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetEchoesPatches)

        QMetaObject.connectSlotsByName(PresetEchoesPatches)
    # setupUi

    def retranslateUi(self, PresetEchoesPatches):
        PresetEchoesPatches.setWindowTitle(QCoreApplication.translate("PresetEchoesPatches", u"Other", None))
        self.include_menu_mod_check.setText(QCoreApplication.translate("PresetEchoesPatches", u"Include Menu Mod", None))
        self.include_menu_mod_label.setText(QCoreApplication.translate("PresetEchoesPatches", u"<html><head/><body><p>Menu Mod is a pratice tool for Echoes, allowing in-game changes to which items you have and warping to all rooms.</p><p>See the <a href=\"https://www.dropbox.com/s/yhqqafaxfo3l4vn/Echoes%20Menu.7z?file_subpath=%2FEchoes+Menu%2Freadme.txt\"><span style=\" text-decoration: underline; color:#0000ff;\">Menu Mod README</span></a> for more details.</p></body></html>", None))
        self.warp_to_start_check.setText(QCoreApplication.translate("PresetEchoesPatches", u"Add warping to starting location to save stations", None))
        self.warp_to_start_label.setText(QCoreApplication.translate("PresetEchoesPatches", u"<html><head/><body><p>Refusing to save at any Save Station will prompt if you want to warp to the starting location (usually Samus' Ship).</p><p><span style=\" color:#005500;\">Usage of the this option is encouraged for all, as it prevents many softlocks that occurs normally in Echoes.</span></p></body></html>", None))
        self.skip_final_bosses_check.setText(QCoreApplication.translate("PresetEchoesPatches", u"Go directly to credits from Sky Temple Gateway", None))
        self.skip_final_bosses_label.setText(QCoreApplication.translate("PresetEchoesPatches", u"<html><head/><body><p>Change the light beam in Sky Temple Gateway to go directly to the credits, skipping the final bosses.</p><p>This changes the requirements to <span style=\" font-weight:600;\">not need the final bosses</span>, turning certain items optional such as Screw Attack and Spider Ball.</p></body></html>", None))
        self.pickup_style_box.setTitle(QCoreApplication.translate("PresetEchoesPatches", u"Pickup style", None))
        self.pickup_model_combo.setItemText(0, QCoreApplication.translate("PresetEchoesPatches", u"Use correct item model, scan and name", None))
        self.pickup_model_combo.setItemText(1, QCoreApplication.translate("PresetEchoesPatches", u"Use correct scan and name, hide the model", None))
        self.pickup_model_combo.setItemText(2, QCoreApplication.translate("PresetEchoesPatches", u"Use correct name, hide the model and scan", None))
        self.pickup_model_combo.setItemText(3, QCoreApplication.translate("PresetEchoesPatches", u"Hide the model, scan and name", None))

        self.pickup_data_source_label.setText(QCoreApplication.translate("PresetEchoesPatches", u"When hiding some part of the pickup, it's replaced with:", None))
        self.pickup_data_source_combo.setItemText(0, QCoreApplication.translate("PresetEchoesPatches", u"Energy Transfer Module's data", None))
        self.pickup_data_source_combo.setItemText(1, QCoreApplication.translate("PresetEchoesPatches", u"A random item data", None))
        self.pickup_data_source_combo.setItemText(2, QCoreApplication.translate("PresetEchoesPatches", u"The data of the pickup in that place", None))

    # retranslateUi




import contextlib
import dataclasses
import datetime
import json
from unittest.mock import MagicMock, PropertyMock, patch, call

import peewee
import pytest

from randovania.game_description.assignment import PickupTarget
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.resources.pickup_entry import PickupEntry, ConditionalResources
from randovania.interface_common.cosmetic_patches import CosmeticPatches
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.preset_migration import VersionedPreset
from randovania.network_common.admin_actions import SessionAdminUserAction, SessionAdminGlobalAction
from randovania.network_common.error import InvalidAction
from randovania.network_common.session_state import GameSessionState
from randovania.server import game_session, database


@pytest.fixture(name="mock_emit_session_update")
def _mock_emit_session_update(mocker) -> MagicMock:
    return mocker.patch("randovania.server.game_session._emit_session_update", autospec=True)


def test_setup_app():
    game_session.setup_app(MagicMock())


def test_game_session_defaults_to_now(clean_database):
    someone = database.User.create(name="Someone")
    database.GameSession.create(name="Debug", num_teams=1, creator=someone)

    session: database.GameSession = database.GameSession.get_by_id(1)
    assert (datetime.datetime.now(datetime.timezone.utc) - session.creation_datetime) < datetime.timedelta(seconds=5)


def test_list_game_sessions(clean_database):
    # Setup
    utc = datetime.timezone.utc
    someone = database.User.create(name="Someone")
    database.GameSession.create(name="Debug", num_teams=1, creator=someone,
                                creation_date=datetime.datetime(2020, 10, 2, 10, 20, tzinfo=utc))
    database.GameSession.create(name="Other", num_teams=2, creator=someone,
                                creation_date=datetime.datetime(2020, 1, 20, 5, 2, tzinfo=utc))
    state = GameSessionState.SETUP.value

    # Run
    result = game_session.list_game_sessions(MagicMock())

    # Assert
    assert result == [
        {'has_password': False, 'id': 1, 'state': state, 'name': 'Debug', 'num_players': 0, 'creator': 'Someone',
         'creation_date': '2020-10-02T10:20:00+00:00'},
        {'has_password': False, 'id': 2, 'state': state, 'name': 'Other', 'num_players': 0, 'creator': 'Someone',
         'creation_date': '2020-01-20T05:02:00+00:00'},
    ]


def test_create_game_session(clean_database, preset_manager):
    # Setup
    user = database.User.create(id=1234, discord_id=5678, name="The Name")
    sio = MagicMock()
    sio.get_current_user.return_value = user

    # Run
    result = game_session.create_game_session(sio, "My Room")

    # Assert
    session = database.GameSession.get(1)
    assert session.name == "My Room"
    assert result == {
        'id': 1,
        'state': GameSessionState.SETUP.value,
        'name': 'My Room',
        'players': [{'admin': True, 'id': 1234, 'name': 'The Name', 'row': 0, 'inventory': None,
                     'connection_state': 'Online, Unknown'}],
        'presets': [preset_manager.default_preset.as_json],
        'actions': [],
        'seed_hash': None,
        'spoiler': None,
        'word_hash': None,
        'permalink': None,
        'generation_in_progress': None,
    }


def test_join_game_session(mock_emit_session_update: MagicMock,
                           clean_database):
    # Setup
    user1 = database.User.create(id=1234, name="The Name")
    user2 = database.User.create(id=1235, name="Other Name")
    sio = MagicMock()
    sio.get_current_user.return_value = user1

    session = database.GameSession.create(name="The Session", password=None, creator=user1)
    database.GameSessionPreset.create(session=session, row=0, preset="{}")
    database.GameSessionMembership.create(user=user2, session=session, row=0, admin=True,
                                          connection_state="Online, Badass")

    # Run
    result = game_session.join_game_session(sio, 1, None)

    # Assert
    mock_emit_session_update.assert_called_once_with(session)
    assert result == {
        'id': 1,
        'state': GameSessionState.SETUP.value,
        'name': 'The Session',
        'players': [
            {'admin': True, 'id': 1235, 'name': 'Other Name', 'row': 0, 'inventory': None,
             'connection_state': 'Online, Badass'},
            {'admin': False, 'id': 1234, 'name': 'The Name', 'row': None, 'inventory': None,
             'connection_state': 'Online, Unknown'},
        ],
        'actions': [],
        'presets': [{}],
        'seed_hash': None,
        'spoiler': None,
        'word_hash': None,
        'permalink': None,
        'generation_in_progress': None,
    }


def test_game_session_request_pickups_missing_membership(clean_database):
    with pytest.raises(peewee.DoesNotExist):
        game_session.game_session_request_pickups(MagicMock(), 1)


def test_game_session_request_pickups_not_in_game(flask_app, clean_database):
    # Setup
    user = database.User.create(id=1234, discord_id=5678, name="The Name")
    session = database.GameSession.create(name="Debug", creator=user)
    database.GameSessionMembership.create(user=user, session=session, row=0, admin=False)

    sio = MagicMock()
    sio.get_current_user.return_value = user

    # Run
    result = game_session.game_session_request_pickups(sio, 1)

    # Assert
    assert result == []


def test_game_session_request_pickups_observer(flask_app, clean_database):
    # Setup
    user = database.User.create(id=1234, discord_id=5678, name="The Name")
    session = database.GameSession.create(name="Debug", creator=user, state=GameSessionState.IN_PROGRESS)
    database.GameSessionMembership.create(user=user, session=session, row=None, admin=False)

    sio = MagicMock()
    sio.get_current_user.return_value = user

    # Run
    result = game_session.game_session_request_pickups(sio, 1)

    # Assert
    assert result == []


@pytest.fixture(name="two_player_session")
def two_player_session_fixture(clean_database):
    user1 = database.User.create(id=1234, name="The Name")
    user2 = database.User.create(id=1235, name="Other Name")

    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.IN_PROGRESS, creator=user1)
    database.GameSessionPreset.create(session=session, row=0, preset="{}")
    database.GameSessionPreset.create(session=session, row=1, preset="{}")

    database.GameSessionMembership.create(user=user1, session=session, row=0, admin=False)
    database.GameSessionMembership.create(user=user2, session=session, row=1, admin=False)
    database.GameSessionTeamAction.create(session=session, provider_row=1, provider_location_index=0, receiver_row=0)

    return session


@patch("randovania.server.game_session._get_pickup_target", autospec=True)
@patch("randovania.server.game_session._get_resource_database", autospec=True)
@patch("randovania.server.database.GameSession.layout_description", new_callable=PropertyMock)
def test_game_session_request_pickups_one_action(mock_session_description: PropertyMock,
                                                 mock_get_resource_database: MagicMock,
                                                 mock_get_pickup_target: MagicMock,
                                                 flask_app, two_player_session, echoes_resource_database):
    # Setup
    sio = MagicMock()
    sio.get_current_user.return_value = database.User.get_by_id(1234)

    pickup = PickupEntry("A", 1, ItemCategory.TEMPLE_KEY, ItemCategory.KEY,
                         (
                             ConditionalResources(None, None, ((echoes_resource_database.item[0], 1),)),
                         ))
    mock_get_pickup_target.return_value = PickupTarget(pickup=pickup, player=0)
    mock_get_resource_database.return_value = echoes_resource_database

    # Run
    result = game_session.game_session_request_pickups(sio, 1)

    # Assert
    mock_get_resource_database.assert_called_once_with(mock_session_description.return_value, 0)
    mock_get_pickup_target.assert_called_once_with(mock_session_description.return_value, 1, 0)
    assert result == [{'provider_name': 'Other Name', 'pickup': '6c8P51ON~'}]


@patch("flask_socketio.emit", autospec=True)
@patch("randovania.server.game_session._get_pickup_target", autospec=True)
@patch("randovania.server.game_session._get_resource_database", autospec=True)
@patch("randovania.server.database.GameSession.layout_description", new_callable=PropertyMock)
def test_game_session_collect_pickup_for_self(mock_session_description: PropertyMock,
                                              mock_get_resource_database: MagicMock,
                                              mock_get_pickup_target: MagicMock,
                                              mock_emit: MagicMock,
                                              flask_app, two_player_session, echoes_resource_database):
    sio = MagicMock()
    sio.get_current_user.return_value = database.User.get_by_id(1234)

    pickup = PickupEntry("A", 1, ItemCategory.TEMPLE_KEY, ItemCategory.KEY,
                         (
                             ConditionalResources(None, None, ((echoes_resource_database.item[0], 1),)),
                         ))
    mock_get_resource_database.return_value = echoes_resource_database
    mock_get_pickup_target.return_value = PickupTarget(pickup, 0)

    # Run
    with flask_app.test_request_context():
        result = game_session.game_session_collect_locations(sio, 1, (0,))

    # Assert
    assert result is None
    mock_emit.assert_not_called()
    mock_get_pickup_target.assert_called_once_with(mock_session_description.return_value, 0, 0)
    with pytest.raises(peewee.DoesNotExist):
        database.GameSessionTeamAction.get(session=two_player_session, provider_row=0,
                                           provider_location_index=0)


@patch("flask_socketio.emit", autospec=True)
@patch("randovania.server.game_session._get_pickup_target", autospec=True)
@patch("randovania.server.database.GameSession.layout_description", new_callable=PropertyMock)
def test_game_session_collect_pickup_etm(mock_session_description: PropertyMock,
                                         mock_get_pickup_target: MagicMock,
                                         mock_emit: MagicMock,
                                         flask_app, two_player_session, echoes_resource_database):
    sio = MagicMock()
    sio.get_current_user.return_value = database.User.get_by_id(1234)

    mock_get_pickup_target.return_value = None

    # Run
    with flask_app.test_request_context():
        result = game_session.game_session_collect_locations(sio, 1, (0,))

    # Assert
    assert result is None
    mock_emit.assert_not_called()
    mock_get_pickup_target.assert_called_once_with(mock_session_description.return_value, 0, 0)
    with pytest.raises(peewee.DoesNotExist):
        database.GameSessionTeamAction.get(session=two_player_session, provider_row=0,
                                           provider_location_index=0)


@pytest.mark.parametrize(("locations_to_collect", "exists"), [
    ((0,), ()),
    ((0,), (0,)),
    ((0, 1), ()),
    ((0, 1), (0,)),
    ((0, 1), (0, 1)),
])
def test_game_session_collect_pickup_other(flask_app, two_player_session, echoes_resource_database,
                                           locations_to_collect, exists, mock_emit_session_update, mocker):
    mock_emit: MagicMock = mocker.patch("flask_socketio.emit", autospec=True)
    mock_get_pickup_target: MagicMock = mocker.patch("randovania.server.game_session._get_pickup_target", autospec=True)
    mock_session_description: PropertyMock = mocker.patch("randovania.server.database.GameSession.layout_description",
                                                          new_callable=PropertyMock)

    sio = MagicMock()
    sio.get_current_user.return_value = database.User.get_by_id(1234)
    mock_get_pickup_target.return_value = PickupTarget(MagicMock(), 1)

    for existing_id in exists:
        database.GameSessionTeamAction.create(session=two_player_session, provider_row=0,
                                              provider_location_index=existing_id, receiver_row=0)

    # Run
    with flask_app.test_request_context():
        result = game_session.game_session_collect_locations(sio, 1, locations_to_collect)

    # Assert
    assert result is None
    mock_get_pickup_target.assert_has_calls([
        call(mock_session_description.return_value, 0, location)
        for location in locations_to_collect
    ])
    for location in locations_to_collect:
        database.GameSessionTeamAction.get(session=two_player_session, provider_row=0,
                                           provider_location_index=location)
    if exists == locations_to_collect:
        mock_emit.assert_not_called()
        mock_emit_session_update.assert_not_called()
    else:
        mock_emit.assert_called_once_with("game_has_update", {"session": 1, "row": 1, },
                                          room=f"game-session-1-1235")
        mock_emit_session_update.assert_called_once_with(database.GameSession.get(id=1))


@pytest.mark.parametrize("is_observer", [False, True])
def test_game_session_admin_player_switch_is_observer(clean_database, flask_app, mock_emit_session_update, is_observer):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.IN_PROGRESS, creator=user1)
    database.GameSessionPreset.create(session=session, row=0, preset="{}")
    database.GameSessionMembership.create(user=user1, session=session, row=None if is_observer else 0, admin=False)
    sio = MagicMock()
    sio.get_current_user.return_value = user1

    # Run
    with flask_app.test_request_context():
        game_session.game_session_admin_player(sio, 1, 1234, SessionAdminUserAction.SWITCH_IS_OBSERVER.value, None)

    # Assert
    membership = database.GameSessionMembership.get(user=user1, session=session)
    assert membership.is_observer != is_observer
    if is_observer:
        assert membership.row == 0
    mock_emit_session_update.assert_called_once_with(database.GameSession.get(id=1))


def test_game_session_admin_player_include_in_session(clean_database, flask_app, mock_emit_session_update):
    users = [
        database.User.create(id=1234 + 1000 * i, name=f"The {i}")
        for i in range(4)
    ]
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.IN_PROGRESS, creator=users[0])
    for i in range(4):
        database.GameSessionPreset.create(session=session, row=i, preset="{}")

    database.GameSessionMembership.create(user=users[0], session=session, row=0, admin=True)
    database.GameSessionMembership.create(user=users[1], session=session, row=2, admin=False)
    database.GameSessionMembership.create(user=users[2], session=session, row=1, admin=False)
    database.GameSessionMembership.create(user=users[3], session=session, row=None, admin=False)
    sio = MagicMock()
    sio.get_current_user.return_value = users[0]

    # Run
    with flask_app.test_request_context():
        game_session.game_session_admin_player(sio, 1, 4234, SessionAdminUserAction.SWITCH_IS_OBSERVER.value, None)

    # Assert
    membership = database.GameSessionMembership.get(user=users[3], session=session)
    assert not membership.is_observer
    assert membership.row == 3
    mock_emit_session_update.assert_called_once_with(database.GameSession.get(id=1))


def test_game_session_admin_kick_last(clean_database, flask_app, mocker):
    mock_emit = mocker.patch("flask_socketio.emit")

    user = database.User.create(id=1234, discord_id=5678, name="The Name")
    sio = MagicMock()
    sio.get_current_user.return_value = user
    game_session.create_game_session(sio, "My Room")
    session = database.GameSession.get_by_id(1)
    database.GameSessionTeamAction.create(session=session, provider_row=0, provider_location_index=0, receiver_row=0,
                                          time=datetime.datetime(2020, 5, 2, 10, 20, tzinfo=datetime.timezone.utc))

    # Run
    with flask_app.test_request_context():
        game_session.game_session_admin_player(sio, 1, 1234, SessionAdminUserAction.KICK.value, None)

    # Assert
    for table in [database.GameSession, database.GameSessionPreset,
                  database.GameSessionMembership, database.GameSessionTeamAction]:
        assert list(table.select()) == []
    assert database.User.get_by_id(1234) == user

    mock_emit.assert_called_once_with(
        'game_session_update',
        {'id': 1, 'name': 'My Room', 'state': 'setup', 'players': [], 'presets': [], 'actions': [],
         'spoiler': None, 'word_hash': None, 'seed_hash': None, 'permalink': None, 'generation_in_progress': None},
        room='game-session-1')


@pytest.mark.parametrize("offset", [-1, 1])
def test_game_session_admin_player_move(clean_database, flask_app, mock_emit_session_update, offset: int):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.IN_PROGRESS, creator=user1)
    database.GameSessionPreset.create(session=session, row=0, preset="{}")
    database.GameSessionPreset.create(session=session, row=1, preset="{}")
    database.GameSessionPreset.create(session=session, row=2, preset="{}")
    database.GameSessionMembership.create(user=user1, session=session, row=1, admin=False)
    sio = MagicMock()
    sio.get_current_user.return_value = user1

    # Run
    with flask_app.test_request_context():
        game_session.game_session_admin_player(sio, 1, 1234, "move", offset)

    # Assert
    membership = database.GameSessionMembership.get(user=user1, session=session)
    assert membership.row == 1 + offset
    mock_emit_session_update.assert_called_once_with(database.GameSession.get(id=1))


@patch("randovania.games.prime.patcher_file.create_patcher_file", autospec=True)
@patch("randovania.server.database.GameSession.layout_description", new_callable=PropertyMock)
def test_game_session_admin_player_patcher_file(mock_layout_description: PropertyMock,
                                                mock_create_patcher_file: MagicMock,
                                                flask_app,
                                                clean_database):
    user1 = database.User.create(id=1234, name="The Name")
    user2 = database.User.create(id=1235, name="Brother")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.IN_PROGRESS, creator=user1)
    database.GameSessionPreset.create(session=session, row=0, preset="{}")
    database.GameSessionPreset.create(session=session, row=1, preset="{}")
    database.GameSessionPreset.create(session=session, row=2, preset="{}")
    database.GameSessionMembership.create(user=user1, session=session, row=2, admin=False)
    database.GameSessionMembership.create(user=user2, session=session, row=1, admin=False)
    sio = MagicMock()
    sio.get_current_user.return_value = user1

    cosmetic = CosmeticPatches(open_map=False)

    # Run
    with flask_app.test_request_context():
        result = game_session.game_session_admin_player(sio, 1, 1234, "create_patcher_file", cosmetic.as_json)

    # Assert
    mock_create_patcher_file.assert_called_once_with(
        mock_layout_description.return_value,
        PlayersConfiguration(2, {
            0: "Player 1",
            1: "Brother",
            2: "The Name",
        }),
        cosmetic
    )
    assert result is mock_create_patcher_file.return_value


def test_game_session_admin_session_delete_session(mock_emit_session_update: MagicMock, flask_app, clean_database):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.SETUP, creator=user1)
    database.GameSessionMembership.create(user=user1, session=session, row=None, admin=True)
    sio = MagicMock()
    sio.get_current_user.return_value = user1

    # Run
    with flask_app.test_request_context():
        game_session.game_session_admin_session(sio, 1, SessionAdminGlobalAction.DELETE_SESSION.value, None)

    # Assert
    mock_emit_session_update.assert_called_once_with(session)
    assert list(database.GameSession.select()) == []


def test_game_session_admin_session_create_row(mock_emit_session_update: MagicMock,
                                               clean_database, flask_app, preset_manager):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.SETUP, creator=user1)
    database.GameSessionMembership.create(user=user1, session=session, row=None, admin=True)
    sio = MagicMock()
    sio.get_current_user.return_value = user1

    # Run
    with flask_app.test_request_context():
        game_session.game_session_admin_session(sio, 1, "create_row", preset_manager.default_preset.as_json)

    # Assert
    mock_emit_session_update.assert_called_once_with(session)
    assert database.GameSession.get_by_id(1).num_rows == 1


def test_game_session_admin_session_change_row(mock_emit_session_update: MagicMock,
                                               clean_database, flask_app, preset_manager):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.SETUP, creator=user1)
    database.GameSessionPreset.create(session=session, row=0, preset="{}")
    database.GameSessionPreset.create(session=session, row=1, preset="{}")
    database.GameSessionMembership.create(user=user1, session=session, row=None, admin=True)
    sio = MagicMock()
    sio.get_current_user.return_value = user1

    # Make sure the preset is using the latest version
    preset_manager.default_preset.ensure_converted()

    # Run
    with flask_app.test_request_context():
        game_session.game_session_admin_session(sio, 1, SessionAdminGlobalAction.CHANGE_ROW.value,
                                                (1, preset_manager.default_preset.as_json))

    # Assert
    mock_emit_session_update.assert_called_once_with(session)
    new_preset_row = database.GameSessionPreset.get(database.GameSessionPreset.session == session,
                                                    database.GameSessionPreset.row == 1)
    assert json.loads(new_preset_row.preset) == preset_manager.default_preset.as_json


def test_game_session_admin_session_delete_row(mock_emit_session_update: MagicMock,
                                               clean_database, flask_app, preset_manager):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.SETUP, creator=user1)
    database.GameSessionPreset.create(session=session, row=0, preset="{}")
    database.GameSessionPreset.create(session=session, row=1, preset="{}")
    database.GameSessionMembership.create(user=user1, session=session, row=None, admin=True)
    sio = MagicMock()
    sio.get_current_user.return_value = user1

    # Run
    with flask_app.test_request_context():
        game_session.game_session_admin_session(sio, 1, SessionAdminGlobalAction.DELETE_ROW.value, 1)

    # Assert
    mock_emit_session_update.assert_called_once_with(session)
    assert database.GameSession.get_by_id(1).num_rows == 1


@pytest.mark.parametrize("not_last_row", [False, True])
def test_game_session_admin_session_delete_row_invalid(mock_emit_session_update,
                                                       clean_database, preset_manager,
                                                       flask_app, not_last_row):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.SETUP, creator=user1)
    database.GameSessionPreset.create(session=session, row=0, preset="{}")
    database.GameSessionMembership.create(user=user1, session=session, row=None, admin=True)
    sio = MagicMock()
    sio.get_current_user.return_value = user1
    if not_last_row:
        database.GameSessionPreset.create(session=session, row=1, preset="{}")
        expected_message = "Can only delete the last row"
        expected_num_rows = 2
    else:
        expected_message = "Can't delete row when there's only one"
        expected_num_rows = 1

    # Run
    with pytest.raises(InvalidAction) as e, flask_app.test_request_context():
        game_session.game_session_admin_session(sio, 1, SessionAdminGlobalAction.DELETE_ROW.value, 0)

    # Assert
    assert e.value.message == expected_message
    mock_emit_session_update.assert_not_called()
    assert database.GameSession.get_by_id(1).num_rows == expected_num_rows


@pytest.mark.parametrize("case", ["to_false", "to_true_free", "to_true_busy"])
def test_game_session_admin_session_update_layout_generation(mock_emit_session_update: MagicMock,
                                                             clean_database, flask_app, case):
    user1 = database.User.create(id=1234, name="The Name")
    user2 = database.User.create(id=1235, name="Other")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.SETUP, creator=user1,
                                          generation_in_progress=user2 if case == "to_true_busy" else None)
    database.GameSessionMembership.create(user=user1, session=session, row=None, admin=True)
    sio = MagicMock()
    sio.get_current_user.return_value = user1

    if case == "to_true_busy":
        expectation = pytest.raises(InvalidAction, match="Generation already in progress by Other.")
        expected_user = user2
    else:
        expectation = contextlib.nullcontext()
        if case == "to_false":
            expected_user = None
        else:
            expected_user = user1

    # Run
    with expectation, flask_app.test_request_context():
        game_session.game_session_admin_session(sio, 1, SessionAdminGlobalAction.UPDATE_LAYOUT_GENERATION.value,
                                                case != "to_false")

    # Assert
    if case == "to_true_busy":
        mock_emit_session_update.assert_not_called()
    else:
        mock_emit_session_update.assert_called_once_with(session)
    assert database.GameSession.get_by_id(1).generation_in_progress == expected_user


def test_game_session_admin_session_change_layout_description(clean_database, preset_manager, mock_emit_session_update,
                                                              mocker, flask_app):
    mock_verify_no_layout_description = mocker.patch("randovania.server.game_session._verify_no_layout_description",
                                                     autospec=True)
    mock_from_json_dict: MagicMock = mocker.patch(
        "randovania.layout.layout_description.LayoutDescription.from_json_dict")

    preset_as_json = json.dumps(preset_manager.default_preset.as_json)
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.SETUP, creator=user1,
                                          generation_in_progress=user1)
    database.GameSessionPreset.create(session=session, row=0, preset=preset_as_json)
    database.GameSessionPreset.create(session=session, row=1, preset=preset_as_json)
    database.GameSessionMembership.create(user=user1, session=session, row=None, admin=True)

    new_preset = preset_manager.default_preset.get_preset()
    new_preset = dataclasses.replace(new_preset,
                                     configuration=dataclasses.replace(new_preset.configuration,
                                                                       menu_mod=False))

    sio = MagicMock()
    sio.get_current_user.return_value = user1
    layout_description = mock_from_json_dict.return_value
    layout_description.as_json = "some_json_string"
    layout_description.permalink.player_count = 2
    layout_description.permalink.presets = {i: new_preset for i in (0, 1)}

    # Run
    with flask_app.test_request_context():
        game_session.game_session_admin_session(sio, 1, SessionAdminGlobalAction.CHANGE_LAYOUT_DESCRIPTION.value,
                                                "layout_description_json")

    # Assert
    mock_emit_session_update.assert_called_once_with(session)
    mock_verify_no_layout_description.assert_called_once_with(session)
    assert database.GameSession.get_by_id(1).layout_description_json == '"some_json_string"'
    assert database.GameSession.get_by_id(1).generation_in_progress is None

    new_session = database.GameSession.get_by_id(1)
    new_json = json.dumps(VersionedPreset.with_preset(new_preset).as_json)
    assert [preset.preset for preset in new_session.presets] == [new_json] * 2


def test_game_session_admin_session_remove_layout_description(mock_emit_session_update: MagicMock, clean_database,
                                                              flask_app):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.SETUP, creator=user1,
                                          generation_in_progress=user1,
                                          layout_description_json="layout_description_json")
    database.GameSessionMembership.create(user=user1, session=session, row=None, admin=True)
    sio = MagicMock()
    sio.get_current_user.return_value = user1

    # Run
    with flask_app.test_request_context():
        game_session.game_session_admin_session(sio, 1, SessionAdminGlobalAction.CHANGE_LAYOUT_DESCRIPTION.value,
                                                None)

    # Assert
    mock_emit_session_update.assert_called_once_with(session)
    assert database.GameSession.get_by_id(1).layout_description_json is None
    assert database.GameSession.get_by_id(1).generation_in_progress is None


@pytest.mark.parametrize("other_user", [False, True])
def test_game_session_admin_session_change_layout_description_invalid(mock_emit_session_update: MagicMock,
                                                                      clean_database, other_user, flask_app):
    user1 = database.User.create(id=1234, name="The Name")
    user2 = database.User.create(id=1235, name="Other")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.SETUP, creator=user1,
                                          generation_in_progress=user2 if other_user else None)
    database.GameSessionMembership.create(user=user1, session=session, row=None, admin=True)
    sio = MagicMock()
    sio.get_current_user.return_value = user1

    if other_user:
        expected_message = "Waiting for a layout from Other."
    else:
        expected_message = "Not waiting for a layout."

    # Run
    with pytest.raises(InvalidAction, match=expected_message), flask_app.test_request_context():
        game_session.game_session_admin_session(sio, 1, SessionAdminGlobalAction.CHANGE_LAYOUT_DESCRIPTION.value,
                                                "layout_description_json")

    # Assert
    mock_emit_session_update.assert_not_called()
    assert database.GameSession.get_by_id(1).layout_description_json is None


@patch("randovania.server.database.GameSession.layout_description", new_callable=PropertyMock)
def test_game_session_admin_session_download_layout_description(mock_layout_description: PropertyMock, flask_app,
                                                                clean_database, mock_emit_session_update):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.SETUP, creator=user1,
                                          layout_description_json="layout_description_json")
    database.GameSessionMembership.create(user=user1, session=session, row=None, admin=False)
    sio = MagicMock()
    sio.get_current_user.return_value = user1

    # Run
    with flask_app.test_request_context():
        result = game_session.game_session_admin_session(sio, 1,
                                                         SessionAdminGlobalAction.DOWNLOAD_LAYOUT_DESCRIPTION.value,
                                                         None)

    # Assert
    mock_emit_session_update.assert_not_called()
    mock_layout_description.assert_called_once()
    assert result == database.GameSession.get_by_id(1).layout_description_json


@patch("randovania.server.database.GameSession.layout_description", new_callable=PropertyMock)
def test_game_session_admin_session_download_layout_description_no_spoiler(mock_layout_description: PropertyMock,
                                                                           clean_database, mock_emit_session_update,
                                                                           flask_app):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.SETUP, creator=user1,
                                          layout_description_json="layout_description_json")
    database.GameSessionMembership.create(user=user1, session=session, row=None, admin=False)
    sio = MagicMock()
    sio.get_current_user.return_value = user1
    mock_layout_description.return_value.permalink.spoiler = False

    # Run
    with pytest.raises(InvalidAction), flask_app.test_request_context():
        game_session.game_session_admin_session(sio, 1, SessionAdminGlobalAction.DOWNLOAD_LAYOUT_DESCRIPTION.value,
                                                None)

    # Assert
    mock_emit_session_update.assert_not_called()
    mock_layout_description.assert_called_once()


@patch("randovania.server.database.GameSession.layout_description", new_callable=PropertyMock)
def test_game_session_admin_session_start_session(mock_session_description: PropertyMock,
                                                  mock_emit_session_update,
                                                  clean_database, preset_manager,
                                                  flask_app):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.SETUP, creator=user1,
                                          layout_description_json="{}")
    database.GameSessionPreset.create(session=session, row=0, preset="{}")
    database.GameSessionMembership.create(user=user1, session=session, row=0, admin=True)
    sio = MagicMock()
    sio.get_current_user.return_value = user1

    # Run
    with flask_app.test_request_context():
        game_session.game_session_admin_session(sio, 1, SessionAdminGlobalAction.START_SESSION.value, None)

    # Assert
    mock_emit_session_update.assert_called_once_with(session)
    assert database.GameSession.get_by_id(1).state == GameSessionState.IN_PROGRESS


@pytest.mark.parametrize("starting_state", [GameSessionState.SETUP, GameSessionState.IN_PROGRESS,
                                            GameSessionState.FINISHED])
def test_game_session_admin_session_finish_session(clean_database, mock_emit_session_update, starting_state,
                                                   flask_app):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=starting_state, creator=user1)
    database.GameSessionMembership.create(user=user1, session=session, row=0, admin=True)
    sio = MagicMock()
    sio.get_current_user.return_value = user1
    if starting_state != GameSessionState.IN_PROGRESS:
        expectation = pytest.raises(InvalidAction, match="Invalid Action: Session is not in progress")
    else:
        expectation = contextlib.nullcontext()

    # Run
    with expectation, flask_app.test_request_context():
        game_session.game_session_admin_session(sio, 1, SessionAdminGlobalAction.FINISH_SESSION.value, None)

    # Assert
    if starting_state != GameSessionState.IN_PROGRESS:
        mock_emit_session_update.assert_not_called()
        assert database.GameSession.get_by_id(1).state == starting_state
    else:
        mock_emit_session_update.assert_called_once_with(session)
        assert database.GameSession.get_by_id(1).state == GameSessionState.FINISHED


def test_game_session_admin_session_reset_session(clean_database, flask_app):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.SETUP, creator=user1)
    database.GameSessionMembership.create(user=user1, session=session, row=0, admin=True)
    sio = MagicMock()
    sio.get_current_user.return_value = user1

    # Run
    with pytest.raises(InvalidAction), flask_app.test_request_context():
        game_session.game_session_admin_session(sio, 1, SessionAdminGlobalAction.RESET_SESSION.value, None)


def test_game_session_admin_session_change_password(clean_database, mock_emit_session_update, flask_app):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.SETUP, creator=user1)
    database.GameSessionMembership.create(user=user1, session=session, row=0, admin=True)
    sio = MagicMock()
    sio.get_current_user.return_value = user1
    expected_password = 'da92cfbc5e318c64e33dc1b0501e5db214cea0e2a5cecabf90269f32f8eaa15f'

    # Run
    with flask_app.test_request_context():
        game_session.game_session_admin_session(sio, 1, SessionAdminGlobalAction.CHANGE_PASSWORD.value, "the_password")

    # Assert
    mock_emit_session_update.assert_called_once_with(session)
    assert database.GameSession.get_by_id(1).password == expected_password


def test_change_row_missing_arguments(flask_app):
    with pytest.raises(InvalidAction), flask_app.test_request_context():
        game_session._change_row(MagicMock(), MagicMock(), (5,))


def test_verify_in_setup(clean_database, flask_app):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.IN_PROGRESS, creator=user1,
                                          layout_description_json="{}")

    with pytest.raises(InvalidAction), flask_app.test_request_context():
        game_session._verify_in_setup(session)


def test_verify_no_layout_description(clean_database, flask_app):
    user1 = database.User.create(id=1234, name="The Name")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.IN_PROGRESS, creator=user1,
                                          layout_description_json="{}")

    with pytest.raises(InvalidAction), flask_app.test_request_context():
        game_session._verify_in_setup(session)


def test_game_session_request_update(clean_database, mocker, flask_app):
    mock_layout = mocker.patch("randovania.server.database.GameSession.layout_description", new_callable=PropertyMock)
    target = mock_layout.return_value.all_patches.__getitem__.return_value.pickup_assignment.__getitem__.return_value
    target.pickup.name = "The Pickup"
    mock_layout.return_value.shareable_word_hash = "Words of O-Lir"
    mock_layout.return_value.shareable_hash = "ABCDEFG"
    mock_layout.return_value.permalink.spoiler = True
    mock_layout.return_value.permalink.as_base64_str = "<permalink>"

    user1 = database.User.create(id=1234, name="The Name")
    user2 = database.User.create(id=1235, name="Other")
    session = database.GameSession.create(id=1, name="Debug", state=GameSessionState.IN_PROGRESS, creator=user1,
                                          layout_description_json="{}")
    database.GameSessionMembership.create(user=user1, session=session, row=0, admin=True,
                                          connection_state="Something")
    database.GameSessionMembership.create(user=user2, session=session, row=1, admin=False,
                                          connection_state="Game")
    database.GameSessionTeamAction.create(session=session, provider_row=1, provider_location_index=0, receiver_row=0,
                                          time=datetime.datetime(2020, 5, 2, 10, 20, tzinfo=datetime.timezone.utc))

    # Run
    with flask_app.test_request_context():
        result = game_session.game_session_request_update(MagicMock(), 1)

    # Assert
    assert result == {
        "id": 1,
        "name": "Debug",
        "state": GameSessionState.IN_PROGRESS.value,
        "players": [
            {
                "id": 1234,
                "name": "The Name",
                "row": 0,
                "admin": True,
                'inventory': None,
                'connection_state': 'Something',
            },
            {
                "id": 1235,
                "name": "Other",
                "row": 1,
                "admin": False,
                'inventory': None,
                'connection_state': 'Game',
            },
        ],
        "presets": [],
        "actions": [
            {
                "message": "Other found The Pickup for The Name.",
                "time": "2020-05-02T10:20:00+00:00",
            }
        ],
        "spoiler": True,
        "word_hash": "Words of O-Lir",
        "seed_hash": "ABCDEFG",
        "permalink": "<permalink>",
        "generation_in_progress": None,
    }



import dataclasses
import json
import logging
from typing import Dict, List, Union

import PySide2
from PySide2.QtCore import QTimer, Signal, Qt
from PySide2.QtGui import QPixmap
from PySide2.QtWidgets import QMainWindow, QLabel
from asyncqt import asyncSlot

from randovania import get_data_path
from randovania.game_connection.connection_base import GameConnectionStatus, InventoryItem
from randovania.game_connection.game_connection import GameConnection
from randovania.game_description import default_database
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.resource_database import find_resource_info_with_long_name
from randovania.games.game import RandovaniaGame
from randovania.gui.generated.auto_tracker_window_ui import Ui_AutoTrackerWindow
from randovania.gui.lib import common_qt_lib
from randovania.gui.lib.clickable_label import ClickableLabel
from randovania.gui.lib.game_connection_setup import GameConnectionSetup
from randovania.gui.lib.pixmap_lib import paint_with_opacity
from randovania.interface_common.options import Options


@dataclasses.dataclass(frozen=True)
class Element:
    label: Union[QLabel, ClickableLabel]
    resources: List[ItemResourceInfo]
    text_template: str


class AutoTrackerWindow(QMainWindow, Ui_AutoTrackerWindow):
    _hooked = False
    _base_address: int
    give_item_signal = Signal(PickupEntry)

    def __init__(self, game_connection: GameConnection, options: Options):
        super().__init__()
        self.setupUi(self)
        self.game_connection = game_connection
        common_qt_lib.set_default_window_icon(self)

        self._tracker_elements: List[Element] = []
        self.create_tracker()

        self.game_connection_setup = GameConnectionSetup(self, self.game_connection_tool, self.connection_status_label,
                                                         self.game_connection, options)
        self.force_update_button.clicked.connect(self.on_force_update_button)

        self._update_timer = QTimer(self)
        self._update_timer.setInterval(100)
        self._update_timer.timeout.connect(self._on_timer_update)
        self._update_timer.setSingleShot(True)

    def showEvent(self, event: PySide2.QtGui.QShowEvent):
        self._update_timer.start()
        super().showEvent(event)

    def hideEvent(self, event: PySide2.QtGui.QHideEvent):
        self._update_timer.stop()
        super().hideEvent(event)

    def _update_tracker_from_hook(self, inventory: Dict[ItemResourceInfo, InventoryItem]):
        for element in self._tracker_elements:
            amount = 0
            capacity = 0
            max_capacity = 0
            for resource in element.resources:
                current = inventory.get(resource, InventoryItem(0, 0))
                amount += current.amount
                capacity += current.capacity
                max_capacity += resource.max_capacity

            if isinstance(element.label, ClickableLabel):
                element.label.set_checked(max_capacity == 0 or capacity > 0)
            else:
                element.label.setText(element.text_template.format(
                    amount=amount,
                    capacity=capacity,
                    max_capacity=max_capacity,
                ))

    @asyncSlot()
    async def _on_timer_update(self):
        try:
            current_status = self.game_connection.current_status
            if current_status == GameConnectionStatus.InGame or current_status == GameConnectionStatus.TrackerOnly:
                inventory = self.game_connection.get_current_inventory()
                self._update_tracker_from_hook(inventory)
        finally:
            self._update_timer.start()

    def delete_tracker(self):
        for element in self._tracker_elements:
            element.label.deleteLater()

        self._tracker_elements.clear()

    def create_tracker(self):
        self.delete_tracker()
        game_enum = RandovaniaGame.PRIME2

        resource_database = default_database.resource_database_for(game_enum)
        with get_data_path().joinpath(f"gui_assets/tracker/{game_enum.value}.json").open("r") as tracker_details_file:
            tracker_details = json.load(tracker_details_file)

        for element in tracker_details["elements"]:
            text_template = ""

            if "image_path" in element:
                image_path = get_data_path().joinpath(element["image_path"])
                pixmap = QPixmap(str(image_path))

                label = ClickableLabel(self.inventory_group, paint_with_opacity(pixmap, 0.3),
                                       paint_with_opacity(pixmap, 1.0))
                label.set_checked(False)
                label.set_ignore_mouse_events(True)

            elif "label" in element:
                label = QLabel(self.inventory_group)
                label.setAlignment(Qt.AlignCenter)
                text_template = element["label"]

            else:
                raise ValueError(f"Invalid element: {element}")

            resources = [
                find_resource_info_with_long_name(resource_database.item, resource_name)
                for resource_name in element["resources"]
            ]
            self._tracker_elements.append(Element(label, resources, text_template))
            self.inventory_layout.addWidget(label, element["row"], element["column"])

    @asyncSlot()
    async def on_force_update_button(self):
        await self.game_connection.backend.update_current_inventory()
        inventory = self.game_connection.get_current_inventory()
        print("Inventory:" + "\n".join(
            f"{item.long_name}: {inv_item.amount}/{inv_item.capacity}"
            for item, inv_item in sorted(inventory.items(), key=lambda it: it[0].long_name)
        ))



import dataclasses
import struct
from pathlib import Path
from typing import Tuple, Optional, BinaryIO, Iterable

from randovania.dol_patching import assembler

_NUM_TEXT_SECTIONS = 7
_NUM_DATA_SECTIONS = 11
_NUM_SECTIONS = _NUM_TEXT_SECTIONS + _NUM_DATA_SECTIONS


@dataclasses.dataclass(frozen=True)
class Section:
    offset: int
    base_address: int
    size: int


@dataclasses.dataclass(frozen=True)
class DolHeader:
    sections: Tuple[Section, ...]
    bss_address: int
    bss_size: int
    entry_point: int

    @classmethod
    def from_bytes(cls, data: bytes) -> "DolHeader":
        struct_format = f">{_NUM_SECTIONS}L"
        offset_for_section = struct.unpack_from(struct_format, data, 0)
        base_address_for_section = struct.unpack_from(struct_format, data, 0x48)
        size_for_section = struct.unpack_from(struct_format, data, 0x90)

        bss_address, bss_size, entry_point = struct.unpack_from(">LLL", data, 0xD8)
        sections = tuple(Section(offset_for_section[i], base_address_for_section[i], size_for_section[i])
                         for i in range(_NUM_SECTIONS))

        return cls(sections, bss_address, bss_size, entry_point)

    def offset_for_address(self, address: int) -> Optional[int]:
        for section in self.sections:
            relative_to_base = address - section.base_address
            if 0 <= relative_to_base < section.size:
                return section.offset + relative_to_base

        return None


class DolFile:
    dol_file: Optional[BinaryIO] = None
    editable: bool = False

    def __init__(self, dol_path: Path):
        with dol_path.open("rb") as f:
            header_bytes = f.read(0x100)

        self.dol_path = dol_path
        self.header = DolHeader.from_bytes(header_bytes)

    def set_editable(self, editable: bool):
        self.editable = editable

    def __enter__(self):
        if self.editable:
            f = self.dol_path.open("r+b")
        else:
            f = self.dol_path.open("rb")
        self.dol_file = f.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.dol_file.__exit__(exc_type, exc_type, exc_tb)
        self.dol_file = None

    def offset_for_address(self, address: int) -> int:
        offset = self.header.offset_for_address(address)
        if offset is None:
            raise ValueError(f"Address 0x{address:x} could not be resolved for dol at {self.dol_file}")
        return offset

    def read(self, address: int, size: int) -> bytes:
        offset = self.offset_for_address(address)
        self.dol_file.seek(offset)
        return self.dol_file.read(size)

    def write(self, address: int, code_points: Iterable[int]):
        offset = self.offset_for_address(address)
        self.dol_file.seek(offset)
        self.dol_file.write(bytes(code_points))

    def write_instructions(self, address: int, instructions: Iterable[assembler.BaseInstruction]):
        self.write(address, assembler.assemble_instructions(address, instructions))



from enum import Enum
from typing import Tuple, Dict

from randovania.bitpacking.bitpacking import BitPackEnum


class ItemCategory(BitPackEnum, Enum):
    VISOR = "visor"
    SUIT = "suit"
    BEAM = "beam"
    MORPH_BALL = "morph_ball"
    MOVEMENT = "movement"
    MISSILE = "missile"
    CHARGE_COMBO = "charge_combo"
    TRANSLATOR = "translator"
    PRIME3_GRAPPLE = "grapple"
    HYPERMODE = "hypermode"
    ENERGY_TANK = "energy_tank"
    TEMPLE_KEY = "temple_key"
    SKY_TEMPLE_KEY = "sky_temple_key"
    ETM = "etm"
    KEY = "key"
    MISSILE_RELATED = "missile_related"
    MORPH_BALL_RELATED = "morph_ball_related"
    BEAM_RELATED = "beam_related"
    LIFE_SUPPORT = "life_support"
    HUD = "hud"
    EXPANSION = "expansion"

    @property
    def is_major_category(self) -> bool:
        return self in MAJOR_ITEM_CATEGORIES

    @property
    def is_key(self) -> bool:
        return self in TEMPLE_KEY_CATEGORIES

    @property
    def long_name(self):
        return LONG_NAMES[self]

    @property
    def hint_details(self) -> Tuple[str, str]:
        return HINT_DETAILS[self]

    @property
    def general_details(self) -> Tuple[str, str]:
        if self.is_major_category:
            return "a ", "major upgrade"
        elif self.is_key:
            return "a ", "key"
        else:
            return "an ", "expansion"


MAJOR_ITEM_CATEGORIES = {
    ItemCategory.VISOR,
    ItemCategory.SUIT,
    ItemCategory.BEAM,
    ItemCategory.MORPH_BALL,
    ItemCategory.MOVEMENT,
    ItemCategory.MISSILE,
    ItemCategory.CHARGE_COMBO,
    ItemCategory.TRANSLATOR,
    ItemCategory.PRIME3_GRAPPLE,
    ItemCategory.HYPERMODE,
}

TEMPLE_KEY_CATEGORIES = {
    ItemCategory.TEMPLE_KEY, ItemCategory.SKY_TEMPLE_KEY
}

LONG_NAMES = {
    ItemCategory.VISOR: "Visors",
    ItemCategory.SUIT: "Suits",
    ItemCategory.BEAM: "Beams",
    ItemCategory.MORPH_BALL: "Morph Ball",
    ItemCategory.MOVEMENT: "Movement",
    ItemCategory.MISSILE: "Missile",
    ItemCategory.CHARGE_COMBO: "Charge Combos",
    ItemCategory.TRANSLATOR: "Translators",
    ItemCategory.PRIME3_GRAPPLE: "Grapple",
    ItemCategory.HYPERMODE: "Hypermode",
    ItemCategory.ENERGY_TANK: "Energy Tanks",
}

HINT_DETAILS: Dict[ItemCategory, Tuple[str, str]] = {
    ItemCategory.VISOR: ("a ", "visor"),
    ItemCategory.SUIT: ("a ", "suit"),
    ItemCategory.BEAM: ("a ", "beam"),
    ItemCategory.MORPH_BALL: ("a ", "morph ball system"),
    ItemCategory.MOVEMENT: ("a ", "movement system"),
    ItemCategory.MISSILE: ("a ", "missile system"),
    ItemCategory.CHARGE_COMBO: ("a ", "charge combo"),
    ItemCategory.TRANSLATOR: ("a ", "translator"),
    ItemCategory.PRIME3_GRAPPLE: ("a ", "grapple system"),
    ItemCategory.HYPERMODE: ("a ", "hypermode system"),
    ItemCategory.ENERGY_TANK: ("an ", "Energy Tank"),
    ItemCategory.TEMPLE_KEY: ("a ", "red Temple Key"),
    ItemCategory.SKY_TEMPLE_KEY: ("a ", "Sky Temple Key"),
    ItemCategory.ETM: ("an ", "Energy Transfer Module"),
    ItemCategory.KEY: ("a ", "key"),
    ItemCategory.MISSILE_RELATED: ("a ", "missile-related upgrade"),
    ItemCategory.MORPH_BALL_RELATED: ("a ", "morph ball-related upgrade"),
    ItemCategory.BEAM_RELATED: ("a ", "beam-related upgrade"),
    ItemCategory.LIFE_SUPPORT: ("a ", "life support system"),
    ItemCategory.HUD: ("a ", "HUD system"),
    ItemCategory.EXPANSION: ("an ", "expansion"),
}



import copy
from typing import Dict, Iterator, Tuple, Set, Callable

import networkx

from randovania.game_description.requirements import Requirement


class BaseGraph:
    def copy(self):
        raise NotImplementedError()

    def add_node(self, node: int):
        raise NotImplementedError()

    def add_edge(self, previous_node: int, next_node: int, requirement: Requirement):
        raise NotImplementedError()

    def remove_edge(self, previous: int, target: int):
        raise NotImplementedError()

    def has_edge(self, previous_node: int, next_node: int) -> bool:
        raise NotImplementedError()

    def __contains__(self, item: int):
        raise NotImplementedError()

    def edges_data(self) -> Iterator[Tuple[int, int, Requirement]]:
        raise NotImplementedError()

    def multi_source_dijkstra(self, sources: Set[int], weight: Callable[[int, int, Requirement], float]):
        raise NotImplementedError()

    def single_source_dijkstra_path(self, source: int):
        raise NotImplementedError()

    def strongly_connected_components(self) -> Iterator[Set[int]]:
        raise NotImplementedError()


class RandovaniaGraph(BaseGraph):
    edges: Dict[int, Dict[int, Requirement]]

    @classmethod
    def new(cls):
        return cls({})

    def __init__(self, edges: Dict[int, Dict[int, Requirement]]):
        self.edges = edges

    def copy(self):
        return RandovaniaGraph(
            {
                source: copy.copy(data)
                for source, data in self.edges.items()
            },
        )

    def add_node(self, node: int):
        if node not in self.edges:
            self.edges[node] = {}

    def add_edge(self, previous_node: int, next_node: int, requirement: Requirement):
        self.edges[previous_node][next_node] = requirement

    def remove_edge(self, previous: int, target: int):
        self.edges[previous].pop(target)

    def has_edge(self, previous_node: int, next_node: int) -> bool:
        return next_node in self.edges.get(previous_node, {})

    def __contains__(self, item: int):
        return item in self.edges

    def edges_data(self):
        for source, data in self.edges.items():
            for target, requirement in data.items():
                yield source, target, requirement

    def multi_source_dijkstra(self, sources: Set[int], weight: Callable[[int, int, Requirement], float]):
        return networkx.multi_source_dijkstra(self, sources, weight=weight)

    def single_source_dijkstra_path(self, source: int):
        return networkx.single_source_dijkstra_path(self, source, weight="unweighted")

    def strongly_connected_components(self) -> Iterator[Set[int]]:
        yield from networkx.strongly_connected_components(self)

    # Networkx compatibility API
    def is_multigraph(self):
        return False

    def is_directed(self):
        return True

    @property
    def _succ(self):
        return self.edges

    def __iter__(self):
        yield from self.edges.keys()

    def __getitem__(self, n):
        return self.edges[n].keys()



from PySide2 import QtWidgets, QtGui
from PySide2.QtCore import Qt

_current_dark_theme = False


def set_dark_theme(active: bool):
    global _current_dark_theme
    if _current_dark_theme == active:
        return

    app: QtWidgets.QApplication = QtWidgets.QApplication.instance()
    new_palette = QtGui.QPalette(app.palette())

    if active:
        import qdarkstyle
        style = qdarkstyle.load_stylesheet(qt_api='pyside2')
        new_palette.setColor(QtGui.QPalette.Link, Qt.cyan)
        new_palette.setColor(QtGui.QPalette.LinkVisited, Qt.cyan)
        style += """
        QGroupBox {
            padding: 0px;
        }
        QGroupBox::title {
          padding-bottom: 12px;
        }
        QComboBox {
            padding-right: 10px;
        }
        QPushButton {
            min-width: 60px;
        }
        QToolButton {
            border: 1px solid #32414B;
        }

            """
    else:
        style = ""
        new_palette.setColor(QtGui.QPalette.Link, Qt.blue)

    _current_dark_theme = active
    app.setStyleSheet(style)
    app.setPalette(new_palette)



import datetime
import functools
import json
from pathlib import Path
from typing import Optional, Set

import pypresence
from PySide2.QtCore import Signal
from PySide2.QtWidgets import QWidget
from cryptography.fernet import Fernet

import randovania
from randovania.gui.lib import async_dialog
from randovania.network_client.game_session import GameSessionEntry, User
from randovania.network_client.network_client import NetworkClient, ConnectionState, UnableToConnect
from randovania.network_common.error import InvalidAction, NotAuthorizedForAction, ServerError


class QtNetworkClient(QWidget, NetworkClient):
    Connect = Signal()
    ConnectError = Signal()
    Disconnect = Signal()
    UserChanged = Signal(User)
    ConnectionStateUpdated = Signal(ConnectionState)
    GameSessionUpdated = Signal(GameSessionEntry)
    GameUpdateNotification = Signal()

    discord: Optional[pypresence.AioClient]

    def __init__(self, user_data_dir: Path):
        super().__init__()
        NetworkClient.__init__(self, user_data_dir.joinpath("network_client"), randovania.get_configuration())
        from randovania.gui.lib import common_qt_lib
        common_qt_lib.set_default_window_icon(self)

        if "discord_client_id" in self.configuration:
            self.discord = pypresence.AioClient(self.configuration["discord_client_id"])
            self.discord._events_on = False  # workaround for broken AioClient
        else:
            self.discord = None

    @NetworkClient.connection_state.setter
    def connection_state(self, value: ConnectionState):
        NetworkClient.connection_state.fset(self, value)
        self.ConnectionStateUpdated.emit(value)

    async def on_connect(self):
        await super().on_connect()
        self.Connect.emit()

    async def on_connect_error(self, error_message: str):
        await super().on_connect_error(error_message)
        self.ConnectError.emit()

    async def on_disconnect(self):
        await super().on_disconnect()
        self.Disconnect.emit()

    async def on_user_session_updated(self, new_session: dict):
        await super().on_user_session_updated(new_session)
        self.UserChanged.emit(self.current_user)

    async def on_game_session_updated(self, data):
        await super().on_game_session_updated(data)
        self.GameSessionUpdated.emit(self._current_game_session)

    async def login_with_discord(self):
        if self.discord is None:
            raise RuntimeError("Missing Discord configuration for Randovania")

        await self.discord.start()
        authorize = await self.discord.authorize(self.configuration["discord_client_id"], ['identify'])

        new_session = await self._emit_with_result("login_with_discord", authorize["data"]["code"])
        await self.on_user_session_updated(new_session)

    async def login_as_guest(self, name: str = "Unknown"):
        if "guest_secret" not in self.configuration:
            raise RuntimeError("Missing guest configuration for Randovania")

        fernet = Fernet(self.configuration["guest_secret"].encode("ascii"))
        login_request = fernet.encrypt(json.dumps({
            "name": name,
            "date": datetime.datetime.now(datetime.timezone.utc).isoformat(),
        }).encode("utf-8"))

        new_session = await self._emit_with_result("login_with_guest", login_request)
        await self.on_user_session_updated(new_session)

    async def logout(self):
        self.session_data_path.unlink()
        self._current_user = None
        self._current_game_session = None
        self.connection_state = ConnectionState.ConnectedNotLogged
        await self._emit_with_result("logout")

    async def on_game_update_notification(self, details):
        self.GameUpdateNotification.emit()

    @property
    def available_login_methods(self) -> Set[str]:
        methods = []
        if "guest_secret" in self.configuration:
            methods.append("guest")
        if self.discord is not None:
            methods.append("discord")
        return set(methods)


def handle_network_errors(fn):
    @functools.wraps(fn)
    async def wrapper(self, *args, **kwargs):
        try:
            return await fn(self, *args, **kwargs)

        except InvalidAction as e:
            await async_dialog.warning(self, "Invalid action", f"{e}")

        except ServerError:
            await async_dialog.warning(self, "Server error",
                                       "An error occurred on the server while processing your request.")

        except NotAuthorizedForAction:
            await async_dialog.warning(self, "Unauthorized",
                                       "You're not authorized to perform that action.")

        except UnableToConnect as e:
            await async_dialog.warning(self, "Connection Error",
                                       f"<b>Unable to connect to the server:</b><br /><br />{e.reason}")

    return wrapper



import pytest
from PySide2.QtCore import Qt
from mock import patch, MagicMock, AsyncMock

from randovania.game_connection.connection_base import GameConnectionStatus, InventoryItem
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.resources.pickup_entry import PickupEntry, ConditionalResources
from randovania.games.prime import dol_patcher
from randovania.gui.debug_backend_window import DebugBackendWindow
from randovania.interface_common.enum_lib import iterate_enum


@pytest.fixture(name="backend")
def debug_backend_window(skip_qtbot):
    return DebugBackendWindow()


@pytest.fixture(name="pickup")
def _pickup(echoes_game_description) -> PickupEntry:
    resource = echoes_game_description.resource_database.energy_tank

    return PickupEntry(
        name="Pickup",
        model_index=0,
        item_category=ItemCategory.MOVEMENT,
        broad_category=ItemCategory.LIFE_SUPPORT,
        resources=(
            ConditionalResources(None, None, ((resource, 2),)),
        ),
    )


@pytest.mark.parametrize("expected_status", iterate_enum(GameConnectionStatus))
def test_current_status(backend, expected_status):
    all_status = list(iterate_enum(GameConnectionStatus))

    backend.current_status_combo.setCurrentIndex(all_status.index(expected_status))
    assert backend.current_status == expected_status


@pytest.mark.asyncio
async def test_display_message(backend):
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]

    message = "Foo"
    await backend._perform_single_memory_operations(backend._write_string_to_game_buffer(message))
    backend._write_memory(backend.patches.cstate_manager_global + 0x2, b"\x01")
    await backend.update(1)
    assert backend.messages_list.findItems(message, Qt.MatchFlag.MatchExactly)


@pytest.mark.asyncio
async def test_update_inventory_label(backend, echoes_resource_database):
    backend._inventory = {
        echoes_resource_database.energy_tank: InventoryItem(4, 4),
        echoes_resource_database.multiworld_magic_item: InventoryItem(0, 2),
    }
    backend._update_inventory_label()
    assert "Energy Tank x 4/4" in backend.inventory_label.text()
    assert "Multiworld Magic Identifier x 0/2" in backend.inventory_label.text()


@pytest.mark.asyncio
@patch("randovania.gui.lib.common_qt_lib.get_network_client", autospec=True)
async def test_setup_locations_combo(mock_get_network_client: MagicMock,
                                     backend):
    # Setup
    patcher_data = {
        "pickups": [
            {"pickup_index": i, "hud_text": f"Item {i}"}
            for i in range(110)
        ]
    }
    mock_get_network_client.return_value.session_admin_player = AsyncMock(return_value=patcher_data)
    backend.patches = backend._expected_patches

    # Run
    await backend._setup_locations_combo()

    # Assert
    mock_get_network_client.assert_called_once_with()
    assert backend.collect_location_combo.count() > 100


def test_clear(backend):
    backend.clear()
    assert backend.messages_list.count() == 0
    assert backend.inventory_label.text() == ""



import dataclasses
import datetime
from typing import List, Dict, Optional

from randovania.layout.preset_migration import VersionedPreset
from randovania.network_common.session_state import GameSessionState


@dataclasses.dataclass(frozen=True)
class GameSessionListEntry:
    id: int
    name: str
    has_password: bool
    state: GameSessionState
    num_players: int
    creator: str
    creation_date: datetime.datetime

    def __post_init__(self):
        tzinfo = self.creation_date.tzinfo
        assert tzinfo is not None and tzinfo.utcoffset(self.creation_date) is not None

    @classmethod
    def from_json(cls, data: dict) -> "GameSessionListEntry":
        data["state"] = GameSessionState(data["state"])
        data["creation_date"] = datetime.datetime.fromisoformat(data["creation_date"])
        return GameSessionListEntry(**data)


@dataclasses.dataclass(frozen=True)
class PlayerSessionEntry:
    id: int
    name: str
    row: Optional[int]
    admin: bool
    connection_state: str

    @classmethod
    def from_json(cls, data) -> "PlayerSessionEntry":
        return PlayerSessionEntry(
            id=data["id"],
            name=data["name"],
            row=data["row"],
            admin=data["admin"],
            connection_state=data["connection_state"],
        )

    @property
    def is_observer(self):
        return self.row is None


@dataclasses.dataclass(frozen=True)
class GameSessionAction:
    message: str
    time: datetime.datetime

    @classmethod
    def from_json(cls, data) -> "GameSessionAction":
        return GameSessionAction(
            message=data["message"],
            time=datetime.datetime.fromisoformat(data["time"]),
        )


@dataclasses.dataclass(frozen=True)
class GameSessionEntry:
    id: int
    name: str
    presets: List[VersionedPreset]
    players: Dict[int, PlayerSessionEntry]
    actions: List[GameSessionAction]
    seed_hash: Optional[str]
    word_hash: Optional[str]
    spoiler: Optional[bool]
    permalink: Optional[str]
    state: GameSessionState
    generation_in_progress: Optional[int]

    @property
    def num_admins(self) -> int:
        return sum(1 for player in self.players.values() if player.admin)

    @property
    def num_rows(self) -> int:
        return len(self.presets)

    @classmethod
    def from_json(cls, data) -> "GameSessionEntry":
        player_entries = [
            PlayerSessionEntry.from_json(player_json)
            for player_json in data["players"]
        ]
        return GameSessionEntry(
            id=data["id"],
            name=data["name"],
            presets=[VersionedPreset(preset_json) for preset_json in data["presets"]],
            players={
                player_entry.id: player_entry
                for player_entry in player_entries
            },
            actions=[GameSessionAction.from_json(item) for item in data["actions"]],
            seed_hash=data["seed_hash"],
            word_hash=data["word_hash"],
            spoiler=data["spoiler"],
            permalink=data["permalink"],
            state=GameSessionState(data["state"]),
            generation_in_progress=data["generation_in_progress"],
        )


@dataclasses.dataclass(frozen=True)
class User:
    id: int
    name: str

    @classmethod
    def from_json(cls, data) -> "User":
        return cls(
            id=data["id"],
            name=data["name"],
        )

    @property
    def as_json(self) -> dict:
        return {
            "id": self.id,
            "name": self.name,
        }



import dataclasses

from randovania.interface_common.echoes_user_preferences import EchoesUserPreferences


@dataclasses.dataclass(frozen=True)
class CosmeticPatches:
    disable_hud_popup: bool = True
    speed_up_credits: bool = True
    open_map: bool = True
    pickup_markers: bool = True
    user_preferences: EchoesUserPreferences = dataclasses.field(default_factory=EchoesUserPreferences)

    @property
    def as_json(self) -> dict:
        return {
            "disable_hud_popup": self.disable_hud_popup,
            "speed_up_credits": self.speed_up_credits,
            "open_map": self.open_map,
            "pickup_markers": self.pickup_markers,
            "user_preferences": self.user_preferences.as_json,
        }

    @classmethod
    def from_json_dict(cls, json_dict: dict) -> "CosmeticPatches":
        return cls(
            disable_hud_popup=json_dict["disable_hud_popup"],
            speed_up_credits=json_dict["speed_up_credits"],
            open_map=json_dict["open_map"],
            pickup_markers=json_dict["pickup_markers"],
            user_preferences=(EchoesUserPreferences.from_json_dict(json_dict["user_preferences"])
                              if "user_preferences" in json_dict else EchoesUserPreferences())
        )

    @classmethod
    def default(cls) -> "CosmeticPatches":
        return cls()



import logging
import multiprocessing
import sys

from randovania import cli

logging.basicConfig(level=logging.WARNING)


def main():
    multiprocessing.freeze_support()
    cli.run_cli(sys.argv)


if __name__ == "__main__":
    main()



from PySide2 import QtWidgets

from randovania.layout.preset import Preset


class PresetTab(QtWidgets.QMainWindow):
    @property
    def tab_title(self):
        return self.windowTitle()

    @property
    def uses_patches_tab(self) -> bool:
        raise NotImplementedError()

    def on_preset_changed(self, preset: Preset):
        raise NotImplementedError()



from randovania.gui.dialog.echoes_user_preferences_dialog import EchoesUserPreferencesDialog
from randovania.interface_common.cosmetic_patches import CosmeticPatches
from randovania.interface_common.echoes_user_preferences import EchoesUserPreferences, SoundMode


def test_change_sound_mode(skip_qtbot):
    preferences = CosmeticPatches(user_preferences=EchoesUserPreferences(sound_mode=SoundMode.MONO))

    dialog = EchoesUserPreferencesDialog(None, preferences)
    skip_qtbot.addWidget(dialog)

    dialog.sound_mode_combo.setCurrentIndex(2)

    assert dialog.preferences == EchoesUserPreferences(sound_mode=SoundMode.SURROUND)


def test_change_sfx_volume(skip_qtbot):
    preferences = CosmeticPatches(user_preferences=EchoesUserPreferences(sfx_volume=15))

    dialog = EchoesUserPreferencesDialog(None, preferences)
    skip_qtbot.addWidget(dialog)

    dialog.sfx_volume_slider.setValue(50)

    assert dialog.preferences == EchoesUserPreferences(sfx_volume=50)


def test_change_hud_lag(skip_qtbot):
    preferences = CosmeticPatches(user_preferences=EchoesUserPreferences(hud_lag=False))

    dialog = EchoesUserPreferencesDialog(None, preferences)
    skip_qtbot.addWidget(dialog)

    dialog.hud_lag_check.setChecked(True)

    assert dialog.preferences == EchoesUserPreferences(hud_lag=True)



from typing import List, Callable, TypeVar, Tuple, Dict

from randovania.game_description.area import Area
from randovania.game_description.area_location import AreaLocation
from randovania.game_description.dock import DockWeakness, DockType, DockWeaknessDatabase, DockConnection
from randovania.game_description.echoes_game_specific import EchoesBeamConfiguration, EchoesGameSpecific
from randovania.game_description.game_description import GameDescription
from randovania.game_description.node import GenericNode, DockNode, TeleporterNode, PickupNode, EventNode, Node, \
    TranslatorGateNode, LogbookNode, LoreType, NodeLocation, PlayerShipNode
from randovania.game_description.requirements import ResourceRequirement, Requirement, \
    RequirementOr, RequirementAnd, RequirementTemplate
from randovania.game_description.resources.damage_resource_info import DamageReduction, DamageResourceInfo
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_database import find_resource_info_with_id, ResourceDatabase, \
    find_resource_info_with_long_name, MissingResource
from randovania.game_description.resources.resource_info import ResourceInfo, ResourceGainTuple
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo
from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo
from randovania.game_description.world import World
from randovania.game_description.world_list import WorldList
from randovania.games.game import RandovaniaGame

X = TypeVar('X')
Y = TypeVar('Y')


def read_array(data: List[Y], item_reader: Callable[[Y], X]) -> List[X]:
    return [item_reader(item) for item in data]


def read_resource_info(data: Dict, resource_type: ResourceType) -> SimpleResourceInfo:
    return SimpleResourceInfo(data["index"], data["long_name"],
                              data["short_name"], resource_type)


def read_item_resource_info(data: Dict) -> ItemResourceInfo:
    return ItemResourceInfo(data["index"], data["long_name"],
                            data["short_name"], data["max_capacity"], data.get("extra"))


def read_trick_resource_info(data: Dict) -> TrickResourceInfo:
    return TrickResourceInfo(data["index"], data["long_name"],
                             data["short_name"], data["description"])


def read_resource_info_array(data: List[Dict], resource_type: ResourceType) -> List[SimpleResourceInfo]:
    return read_array(data, lambda info: read_resource_info(info, resource_type=resource_type))


# Damage

def read_damage_reduction(data: Dict, items: List[ItemResourceInfo]) -> DamageReduction:
    return DamageReduction(find_resource_info_with_id(items, data["index"], ResourceType.ITEM),
                           data["multiplier"])


def read_damage_reductions(data: List[Dict], items: List[ItemResourceInfo]) -> Tuple[DamageReduction, ...]:
    return tuple(read_array(data, lambda info: read_damage_reduction(info, items)))


def read_damage_resource_info(data: Dict, items: List[ItemResourceInfo]) -> DamageResourceInfo:
    return DamageResourceInfo(data["index"], data["long_name"],
                              data["short_name"],
                              read_damage_reductions(data["reductions"], items))


def read_damage_resource_info_array(data: List[Dict], items: List[ItemResourceInfo]) -> List[DamageResourceInfo]:
    return read_array(data, lambda info: read_damage_resource_info(info, items))


# Requirement


def read_resource_requirement(data: Dict, resource_database: ResourceDatabase
                              ) -> ResourceRequirement:
    data = data["data"]
    return ResourceRequirement.with_data(
        resource_database,
        ResourceType(data["type"]), data["index"],
        data["amount"], data["negate"])


def read_requirement_and(data: Dict,
                         resource_database: ResourceDatabase,
                         ) -> RequirementAnd:
    return RequirementAnd([
        read_requirement(item, resource_database)
        for item in data["data"]
    ])


def read_requirement_or(data: Dict,
                        resource_database: ResourceDatabase) -> RequirementOr:
    return RequirementOr([
        read_requirement(item, resource_database)
        for item in data["data"]
    ])


def read_requirement_template(data: Dict, resource_database: ResourceDatabase) -> RequirementTemplate:
    return RequirementTemplate(resource_database, data["data"])


def read_requirement(data: Dict, resource_database: ResourceDatabase) -> Requirement:
    req_type = data["type"]
    if req_type == "resource":
        return read_resource_requirement(data, resource_database)

    elif req_type == "and":
        return read_requirement_and(data, resource_database)

    elif req_type == "or":
        return read_requirement_or(data, resource_database)

    elif req_type == "template":
        return read_requirement_template(data, resource_database)

    else:
        raise ValueError(f"Unknown requirement type: {req_type}")


# Resource Gain

def read_single_resource_gain(item: Dict, database: "ResourceDatabase") -> Tuple[ResourceInfo, int]:
    resource = database.get_by_type_and_index(ResourceType(item["resource_type"]),
                                              item["resource_index"])
    amount = item["amount"]

    return resource, amount


def read_resource_gain_tuple(data: List[Dict], database: "ResourceDatabase") -> ResourceGainTuple:
    return tuple(
        read_single_resource_gain(item, database)
        for item in data
    )


# Dock Weakness

def read_dock_weakness(item: Dict, resource_database: ResourceDatabase, dock_type: DockType) -> DockWeakness:
    return DockWeakness(item["index"],
                        item["name"],
                        item["is_blast_door"],
                        read_requirement(item["requirement"], resource_database),
                        dock_type)


def read_dock_weakness_database(data: Dict,
                                resource_database: ResourceDatabase,
                                ) -> DockWeaknessDatabase:
    door_types = read_array(data["door"], lambda item: read_dock_weakness(item, resource_database, DockType.DOOR))
    portal_types = read_array(data["portal"], lambda item: read_dock_weakness(item, resource_database, DockType.PORTAL))
    morph_ball_types = read_array(data["morph_ball"], lambda item: read_dock_weakness(item, resource_database,
                                                                                      DockType.MORPH_BALL_DOOR))

    return DockWeaknessDatabase(
        door=door_types,
        morph_ball=morph_ball_types,
        other=[
            DockWeakness(0, "Other Door", False, Requirement.trivial(), DockType.OTHER)
        ],
        portal=portal_types)


def read_beam_configuration(data: Dict, resource_database: ResourceDatabase,
                            ) -> EchoesBeamConfiguration:
    return EchoesBeamConfiguration(
        item=resource_database.get_item(data["item_index"]),
        ammo_a=resource_database.get_item(data["ammo_a"]) if data["ammo_a"] is not None else None,
        ammo_b=resource_database.get_item(data["ammo_b"]) if data["ammo_b"] is not None else None,
        uncharged_cost=data["uncharged_cost"],
        charged_cost=data["charged_cost"],
        combo_missile_cost=data["combo_missile_cost"],
        combo_ammo_cost=data["combo_ammo_cost"],
    )


def read_game_specific(data: Dict, resource_database: ResourceDatabase,
                       ) -> EchoesGameSpecific:
    return EchoesGameSpecific(
        energy_per_tank=data["energy_per_tank"],
        safe_zone_heal_per_second=data["safe_zone_heal_per_second"],
        beam_configurations=tuple(
            read_beam_configuration(beam, resource_database)
            for beam in data["beam_configurations"]
        ),
        dangerous_energy_tank=data["dangerous_energy_tank"],
    )


def location_from_json(location: Dict[str, float]) -> NodeLocation:
    return NodeLocation(location["x"], location["y"], location["z"])


class WorldReader:
    resource_database: ResourceDatabase
    dock_weakness_database: DockWeaknessDatabase
    generic_index: int = -1

    def __init__(self,
                 resource_database: ResourceDatabase,
                 dock_weakness_database: DockWeaknessDatabase,
                 ):

        self.resource_database = resource_database
        self.dock_weakness_database = dock_weakness_database

    def _get_scan_visor(self) -> ItemResourceInfo:
        return find_resource_info_with_long_name(
            self.resource_database.item,
            "Scan Visor"
        )

    def _get_command_visor(self) -> ItemResourceInfo:
        return find_resource_info_with_long_name(
            self.resource_database.item,
            "Command Visor"
        )

    def read_node(self, data: Dict) -> Node:
        name: str = data["name"]
        self.generic_index += 1
        try:
            heal: bool = data["heal"]
            node_type: int = data["node_type"]
            location = None
            if data["coordinates"] is not None:
                location = location_from_json(data["coordinates"])

            if node_type == "generic":
                return GenericNode(name, heal, location, self.generic_index)

            elif node_type == "dock":
                return DockNode(name, heal, location, self.generic_index, data["dock_index"],
                                DockConnection(data["connected_area_asset_id"], data["connected_dock_index"]),
                                self.dock_weakness_database.get_by_type_and_index(DockType(data["dock_type"]),
                                                                                  data["dock_weakness_index"]))

            elif node_type == "pickup":
                return PickupNode(name, heal, location, self.generic_index, PickupIndex(data["pickup_index"]),
                                  data["major_location"])

            elif node_type == "teleporter":
                instance_id = data["teleporter_instance_id"]

                destination_world_asset_id = data["destination_world_asset_id"]
                destination_area_asset_id = data["destination_area_asset_id"]

                return TeleporterNode(name, heal, location, self.generic_index, instance_id,
                                      AreaLocation(destination_world_asset_id, destination_area_asset_id),
                                      data["scan_asset_id"],
                                      data["keep_name_when_vanilla"],
                                      data["editable"],
                                      )

            elif node_type == "event":
                return EventNode(name, heal, location, self.generic_index,
                                 self.resource_database.get_by_type_and_index(ResourceType.EVENT, data["event_index"]))

            elif node_type == "translator_gate":
                return TranslatorGateNode(name, heal, location, self.generic_index,
                                          TranslatorGate(data["gate_index"]),
                                          self._get_scan_visor())

            elif node_type == "logbook":
                lore_type = LoreType(data["lore_type"])

                if lore_type == LoreType.LUMINOTH_LORE:
                    required_translator = self.resource_database.get_item(data["extra"])
                else:
                    required_translator = None

                if lore_type in {LoreType.LUMINOTH_WARRIOR, LoreType.SKY_TEMPLE_KEY_HINT}:
                    hint_index = data["extra"]
                else:
                    hint_index = None

                return LogbookNode(name, heal, location, self.generic_index, data["string_asset_id"],
                                   self._get_scan_visor(),
                                   lore_type,
                                   required_translator,
                                   hint_index)

            elif node_type == "player_ship":
                return PlayerShipNode(name, heal, location, self.generic_index,
                                      read_requirement(data["is_unlocked"], self.resource_database),
                                      self._get_command_visor())

            else:
                raise Exception(f"Unknown type: {node_type}")

        except Exception as e:
            raise Exception(f"In node {name}, got error: {e}")

    def read_area(self, data: Dict) -> Area:
        nodes = read_array(data["nodes"], self.read_node)
        nodes_by_name = {node.name: node for node in nodes}

        connections = {}
        for i, origin_data in enumerate(data["nodes"]):
            origin = nodes[i]
            connections[origin] = {}

            for target_name, target_requirement in origin_data["connections"].items():
                try:
                    the_set = read_requirement(target_requirement, self.resource_database)
                except MissingResource as e:
                    raise MissingResource(
                        f"In area {data['name']}, connection from {origin.name} to {target_name} got error: {e}")

                if the_set != Requirement.impossible():
                    connections[origin][nodes_by_name[target_name]] = the_set

        area_name = data["name"]
        try:
            return Area(area_name, data["in_dark_aether"], data["asset_id"], data["default_node_index"],
                        data["valid_starting_location"],
                        nodes, connections)
        except KeyError as e:
            raise KeyError(f"Missing key `{e}` for area `{area_name}`")

    def read_area_list(self, data: List[Dict]) -> List[Area]:
        return read_array(data, self.read_area)

    def read_world(self, data: Dict) -> World:
        return World(data["name"], data["dark_name"], data["asset_id"],
                     self.read_area_list(data["areas"]))

    def read_world_list(self, data: List[Dict]) -> WorldList:
        return WorldList(read_array(data, self.read_world))


def read_requirement_templates(data: Dict, database: ResourceDatabase) -> Dict[str, Requirement]:
    return {
        name: read_requirement(item, database)
        for name, item in data.items()
    }


def read_resource_database(data: Dict) -> ResourceDatabase:
    item = read_array(data["items"], read_item_resource_info)
    db = ResourceDatabase(
        item=item,
        event=read_resource_info_array(data["events"], ResourceType.EVENT),
        trick=read_array(data["tricks"], read_trick_resource_info),
        damage=read_damage_resource_info_array(data["damage"], item),
        version=read_resource_info_array(data["versions"], ResourceType.VERSION),
        misc=read_resource_info_array(data["misc"], ResourceType.MISC),
        requirement_template={},
        energy_tank_item_index=data["energy_tank_item_index"],
        item_percentage_index=data["item_percentage_index"],
        multiworld_magic_item_index=data["multiworld_magic_item_index"]
    )
    db.requirement_template.update(read_requirement_templates(data["requirement_template"], db))
    return db


def read_initial_states(data: Dict[str, List], resource_database: ResourceDatabase) -> Dict[str, ResourceGainTuple]:
    return {
        name: read_resource_gain_tuple(item, resource_database)
        for name, item in data.items()
    }


def decode_data_with_world_reader(data: Dict) -> Tuple[WorldReader, GameDescription]:
    game = RandovaniaGame(data["game"])

    resource_database = read_resource_database(data["resource_database"])
    dock_weakness_database = read_dock_weakness_database(data["dock_weakness_database"], resource_database)
    if game == RandovaniaGame.PRIME2:
        game_specific = read_game_specific(data["game_specific"], resource_database)
    else:
        game_specific = None

    world_reader = WorldReader(resource_database, dock_weakness_database)
    world_list = world_reader.read_world_list(data["worlds"])

    victory_condition = read_requirement(data["victory_condition"], resource_database)
    starting_location = AreaLocation.from_json(data["starting_location"])
    initial_states = read_initial_states(data["initial_states"], resource_database)

    return world_reader, GameDescription(
        game=game,
        resource_database=resource_database,
        game_specific=game_specific,
        dock_weakness_database=dock_weakness_database,
        world_list=world_list,
        victory_condition=victory_condition,
        starting_location=starting_location,
        initial_states=initial_states,
    )


def decode_data(data: Dict) -> GameDescription:
    return decode_data_with_world_reader(data)[1]



from typing import Dict

from randovania.game_description.game_description import GameDescription
from randovania.game_description.node import Node
from randovania.game_description.requirements import RequirementSet
from randovania.layout.echoes_configuration import EchoesConfiguration


class Logic:
    """Extra information that persists even after a backtrack, to prevent irrelevant backtracking."""

    game: GameDescription
    configuration: EchoesConfiguration
    additional_requirements: Dict[Node, RequirementSet]

    def __init__(self, game: GameDescription, configuration: EchoesConfiguration):
        self.game = game
        self.configuration = configuration
        self.additional_requirements = {}

    def get_additional_requirements(self, node: Node) -> RequirementSet:
        return self.additional_requirements.get(node, RequirementSet.trivial())



from randovania.game_description.requirements import RequirementOr, RequirementAnd, ResourceRequirement
from randovania.game_description.resources.resource_type import ResourceType
from randovania.gui.dialog import connections_editor


def test_build_no_changes(skip_qtbot, echoes_resource_database):
    # Setup
    def mk_req(index: int):
        return ResourceRequirement.with_data(echoes_resource_database, ResourceType.ITEM, index, 1, False)

    requirement = RequirementOr([
        RequirementAnd([
            mk_req(1),
            mk_req(2),
        ]),
        mk_req(0),
    ])

    # Run
    editor = connections_editor.ConnectionsEditor(None, echoes_resource_database,
                                                  requirement)
    skip_qtbot.addWidget(editor)
    result = editor.final_requirement

    # Assert
    assert result == requirement



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'auto_tracker_window.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_AutoTrackerWindow(object):
    def setupUi(self, AutoTrackerWindow):
        if not AutoTrackerWindow.objectName():
            AutoTrackerWindow.setObjectName(u"AutoTrackerWindow")
        AutoTrackerWindow.resize(402, 248)
        self.centralwidget = QWidget(AutoTrackerWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.gridLayout = QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.inventory_group = QGroupBox(self.centralwidget)
        self.inventory_group.setObjectName(u"inventory_group")
        self.inventory_group.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.inventory_layout = QGridLayout(self.inventory_group)
        self.inventory_layout.setObjectName(u"inventory_layout")

        self.gridLayout.addWidget(self.inventory_group, 0, 0, 1, 1)

        self.game_connection_layout = QHBoxLayout()
        self.game_connection_layout.setObjectName(u"game_connection_layout")
        self.game_connection_tool = QToolButton(self.centralwidget)
        self.game_connection_tool.setObjectName(u"game_connection_tool")
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.game_connection_tool.sizePolicy().hasHeightForWidth())
        self.game_connection_tool.setSizePolicy(sizePolicy)
        self.game_connection_tool.setPopupMode(QToolButton.InstantPopup)

        self.game_connection_layout.addWidget(self.game_connection_tool)

        self.connection_status_label = QLabel(self.centralwidget)
        self.connection_status_label.setObjectName(u"connection_status_label")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.connection_status_label.sizePolicy().hasHeightForWidth())
        self.connection_status_label.setSizePolicy(sizePolicy1)

        self.game_connection_layout.addWidget(self.connection_status_label)

        self.force_update_button = QPushButton(self.centralwidget)
        self.force_update_button.setObjectName(u"force_update_button")
        sizePolicy.setHeightForWidth(self.force_update_button.sizePolicy().hasHeightForWidth())
        self.force_update_button.setSizePolicy(sizePolicy)

        self.game_connection_layout.addWidget(self.force_update_button)


        self.gridLayout.addLayout(self.game_connection_layout, 1, 0, 1, 1)

        AutoTrackerWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(AutoTrackerWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 402, 21))
        AutoTrackerWindow.setMenuBar(self.menubar)

        self.retranslateUi(AutoTrackerWindow)

        QMetaObject.connectSlotsByName(AutoTrackerWindow)
    # setupUi

    def retranslateUi(self, AutoTrackerWindow):
        AutoTrackerWindow.setWindowTitle(QCoreApplication.translate("AutoTrackerWindow", u"Auto Tracker", None))
        self.inventory_group.setTitle(QCoreApplication.translate("AutoTrackerWindow", u"Inventory", None))
        self.game_connection_tool.setText(QCoreApplication.translate("AutoTrackerWindow", u"Connect to game", None))
        self.connection_status_label.setText(QCoreApplication.translate("AutoTrackerWindow", u"Connection Status", None))
        self.force_update_button.setText(QCoreApplication.translate("AutoTrackerWindow", u"Update", None))
    # retranslateUi




import base64
import hashlib
import json
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from random import Random
from typing import Tuple, Dict

from randovania import get_data_path
from randovania.game_description.game_patches import GamePatches
from randovania.layout import game_patches_serializer
from randovania.layout.permalink import Permalink
from randovania.layout.preset_migration import VersionedPreset


@lru_cache(maxsize=1)
def _shareable_hash_words():
    with (get_data_path() / "hash_words" / "hash_words.json").open() as hash_words_file:
        return json.load(hash_words_file)


CURRENT_DESCRIPTION_SCHEMA_VERSION = 2


def migrate_description(json_dict: dict) -> dict:
    if "schema_version" not in json_dict:
        raise ValueError(f"missing schema_version")

    version = json_dict["schema_version"]
    if version > CURRENT_DESCRIPTION_SCHEMA_VERSION:
        raise ValueError(f"Version {version} is newer than latest supported {CURRENT_DESCRIPTION_SCHEMA_VERSION}")

    if version == 1:
        for game in json_dict["game_modifications"]:
            for hint in game["hints"].values():
                if hint.get("precision") is not None:
                    owner = False
                    if hint["precision"]["item"] == "owner":
                        owner = True
                        hint["precision"]["item"] = "nothing"
                    hint["precision"]["include_owner"] = owner
        version += 1

    json_dict["schema_version"] = version
    return json_dict


@dataclass(frozen=True)
class LayoutDescription:
    version: str
    permalink: Permalink
    all_patches: Dict[int, GamePatches]
    item_order: Tuple[str, ...]

    def __post_init__(self):
        object.__setattr__(self, "__cached_serialized_patches", None)

    @classmethod
    def file_extension(cls) -> str:
        return "rdvgame"

    @classmethod
    def from_json_dict(cls, json_dict: dict) -> "LayoutDescription":
        json_dict = migrate_description(json_dict)

        has_spoiler = "game_modifications" in json_dict
        if not has_spoiler:
            raise ValueError("Unable to read details of seed log with spoiler disabled")

        permalink = Permalink(
            seed_number=json_dict["info"]["seed"],
            spoiler=has_spoiler,
            presets={
                index: VersionedPreset(preset).get_preset()
                for index, preset in enumerate(json_dict["info"]["presets"])
            },
        )

        return LayoutDescription(
            version=json_dict["info"]["version"],
            permalink=permalink,
            all_patches=game_patches_serializer.decode(
                json_dict["game_modifications"], {
                    index: preset.configuration
                    for index, preset in permalink.presets.items()
                }),
            item_order=json_dict["item_order"],
        )

    @classmethod
    def from_file(cls, json_path: Path) -> "LayoutDescription":
        with json_path.open("r") as open_file:
            return cls.from_json_dict(json.load(open_file))

    @property
    def _serialized_patches(self):
        cached_result = object.__getattribute__(self, "__cached_serialized_patches")
        if cached_result is None:
            cached_result = game_patches_serializer.serialize(
                self.all_patches,
                {
                    index: preset.configuration.game_data
                    for index, preset in self.permalink.presets.items()
                })
            object.__setattr__(self, "__cached_serialized_patches", cached_result)

        return cached_result

    @property
    def as_json(self) -> dict:
        result = {
            "schema_version": CURRENT_DESCRIPTION_SCHEMA_VERSION,
            "info": {
                "version": self.version,
                "permalink": self.permalink.as_base64_str,
                "seed": self.permalink.seed_number,
                "presets": [
                    VersionedPreset.with_preset(preset).as_json
                    for preset in self.permalink.presets.values()
                ],
            }
        }

        if self.permalink.spoiler:
            result["game_modifications"] = self._serialized_patches
            result["item_order"] = self.item_order

        return result

    @property
    def _shareable_hash_bytes(self) -> bytes:
        bytes_representation = json.dumps(self._serialized_patches).encode()
        return hashlib.blake2b(bytes_representation, digest_size=5).digest()

    @property
    def shareable_hash(self) -> str:
        return base64.b32encode(self._shareable_hash_bytes).decode()

    @property
    def shareable_word_hash(self) -> str:
        rng = Random(sum([hash_byte * (2 ** 8) ** i for i, hash_byte in enumerate(self._shareable_hash_bytes)]))
        return " ".join(rng.sample(_shareable_hash_words(), 3))

    def save_to_file(self, json_path: Path):
        with json_path.open("w") as open_file:
            json.dump(self.as_json, open_file, indent=4, separators=(',', ': '))

    @property
    def without_item_order(self) -> "LayoutDescription":
        """
        A solver path is way too big to reasonably store for test purposes, so use LayoutDescriptions with an empty one.
        :return:
        """
        return LayoutDescription(
            permalink=self.permalink,
            version=self.version,
            all_patches=self.all_patches,
            item_order=())



from randovania.game_description import node_search


def test_distances_to_node(echoes_game_description):
    world_list = echoes_game_description.world_list
    starting_area = world_list.area_by_area_location(echoes_game_description.starting_location)
    starting_node = world_list.resolve_teleporter_connection(echoes_game_description.starting_location)

    # Run
    result = node_search.distances_to_node(world_list, starting_node)

    # Assert
    assert result[starting_area] == 0



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'seed_details_window.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_SeedDetailsWindow(object):
    def setupUi(self, SeedDetailsWindow):
        if not SeedDetailsWindow.objectName():
            SeedDetailsWindow.setObjectName(u"SeedDetailsWindow")
        SeedDetailsWindow.resize(624, 471)
        self.centralWidget = QWidget(SeedDetailsWindow)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.centralWidget.setLayoutDirection(Qt.LeftToRight)
        self.verticalLayout = QVBoxLayout(self.centralWidget)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.layout_info_tab = QTabWidget(self.centralWidget)
        self.layout_info_tab.setObjectName(u"layout_info_tab")
        self.details_tab = QWidget()
        self.details_tab.setObjectName(u"details_tab")
        self.details_tab_layout = QGridLayout(self.details_tab)
        self.details_tab_layout.setSpacing(6)
        self.details_tab_layout.setContentsMargins(11, 11, 11, 11)
        self.details_tab_layout.setObjectName(u"details_tab_layout")
        self.details_tab_layout.setContentsMargins(4, 8, 4, 0)
        self.customize_user_preferences_button = QPushButton(self.details_tab)
        self.customize_user_preferences_button.setObjectName(u"customize_user_preferences_button")

        self.details_tab_layout.addWidget(self.customize_user_preferences_button, 2, 4, 1, 2)

        self.progress_box = QGroupBox(self.details_tab)
        self.progress_box.setObjectName(u"progress_box")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.progress_box.sizePolicy().hasHeightForWidth())
        self.progress_box.setSizePolicy(sizePolicy)
        self.progress_box_layout = QGridLayout(self.progress_box)
        self.progress_box_layout.setSpacing(6)
        self.progress_box_layout.setContentsMargins(11, 11, 11, 11)
        self.progress_box_layout.setObjectName(u"progress_box_layout")
        self.progress_box_layout.setContentsMargins(2, 4, 2, 4)
        self.stop_background_process_button = QPushButton(self.progress_box)
        self.stop_background_process_button.setObjectName(u"stop_background_process_button")
        self.stop_background_process_button.setEnabled(False)
        sizePolicy1 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.stop_background_process_button.sizePolicy().hasHeightForWidth())
        self.stop_background_process_button.setSizePolicy(sizePolicy1)
        self.stop_background_process_button.setMaximumSize(QSize(75, 16777215))
        self.stop_background_process_button.setCheckable(False)
        self.stop_background_process_button.setFlat(False)

        self.progress_box_layout.addWidget(self.stop_background_process_button, 0, 3, 1, 1)

        self.progress_label = QLabel(self.progress_box)
        self.progress_label.setObjectName(u"progress_label")
        sizePolicy2 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.progress_label.sizePolicy().hasHeightForWidth())
        self.progress_label.setSizePolicy(sizePolicy2)
        font = QFont()
        font.setPointSize(7)
        self.progress_label.setFont(font)
        self.progress_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignVCenter)
        self.progress_label.setWordWrap(True)

        self.progress_box_layout.addWidget(self.progress_label, 0, 2, 1, 1)

        self.progress_bar = QProgressBar(self.progress_box)
        self.progress_bar.setObjectName(u"progress_bar")
        self.progress_bar.setMinimumSize(QSize(150, 0))
        self.progress_bar.setMaximumSize(QSize(150, 16777215))
        self.progress_bar.setValue(0)
        self.progress_bar.setTextVisible(True)
        self.progress_bar.setInvertedAppearance(False)

        self.progress_box_layout.addWidget(self.progress_bar, 0, 0, 1, 2)


        self.details_tab_layout.addWidget(self.progress_box, 4, 0, 1, 7)

        self.layout_description_left_label = QLabel(self.details_tab)
        self.layout_description_left_label.setObjectName(u"layout_description_left_label")
        self.layout_description_left_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.layout_description_left_label.setWordWrap(True)

        self.details_tab_layout.addWidget(self.layout_description_left_label, 3, 0, 1, 3)

        self.layout_description_right_label = QLabel(self.details_tab)
        self.layout_description_right_label.setObjectName(u"layout_description_right_label")
        self.layout_description_right_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.layout_description_right_label.setWordWrap(True)

        self.details_tab_layout.addWidget(self.layout_description_right_label, 3, 3, 1, 4)

        self.player_index_combo = QComboBox(self.details_tab)
        self.player_index_combo.setObjectName(u"player_index_combo")

        self.details_tab_layout.addWidget(self.player_index_combo, 2, 0, 1, 4)

        self.export_iso_button = QPushButton(self.details_tab)
        self.export_iso_button.setObjectName(u"export_iso_button")

        self.details_tab_layout.addWidget(self.export_iso_button, 1, 4, 1, 1)

        self.export_log_button = QPushButton(self.details_tab)
        self.export_log_button.setObjectName(u"export_log_button")

        self.details_tab_layout.addWidget(self.export_log_button, 1, 5, 1, 1)

        self.layout_title_label = QLabel(self.details_tab)
        self.layout_title_label.setObjectName(u"layout_title_label")
        sizePolicy3 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.layout_title_label.sizePolicy().hasHeightForWidth())
        self.layout_title_label.setSizePolicy(sizePolicy3)
        self.layout_title_label.setTextInteractionFlags(Qt.LinksAccessibleByMouse|Qt.TextSelectableByMouse)

        self.details_tab_layout.addWidget(self.layout_title_label, 1, 0, 1, 4)

        self.tool_button = QToolButton(self.details_tab)
        self.tool_button.setObjectName(u"tool_button")
        self.tool_button.setPopupMode(QToolButton.InstantPopup)

        self.details_tab_layout.addWidget(self.tool_button, 1, 6, 1, 1)

        self.permalink_edit = QLineEdit(self.details_tab)
        self.permalink_edit.setObjectName(u"permalink_edit")
        self.permalink_edit.setReadOnly(True)

        self.details_tab_layout.addWidget(self.permalink_edit, 0, 1, 1, 3)

        self.permalink_label = QLabel(self.details_tab)
        self.permalink_label.setObjectName(u"permalink_label")

        self.details_tab_layout.addWidget(self.permalink_label, 0, 0, 1, 1)

        self.layout_info_tab.addTab(self.details_tab, "")
        self.pickup_tab = QWidget()
        self.pickup_tab.setObjectName(u"pickup_tab")
        self.spoiler_pickup_layout = QGridLayout(self.pickup_tab)
        self.spoiler_pickup_layout.setSpacing(6)
        self.spoiler_pickup_layout.setContentsMargins(11, 11, 11, 11)
        self.spoiler_pickup_layout.setObjectName(u"spoiler_pickup_layout")
        self.spoiler_pickup_layout.setContentsMargins(4, 8, 0, 0)
        self.pickup_spoiler_pickup_combobox = QComboBox(self.pickup_tab)
        self.pickup_spoiler_pickup_combobox.addItem("")
        self.pickup_spoiler_pickup_combobox.setObjectName(u"pickup_spoiler_pickup_combobox")

        self.spoiler_pickup_layout.addWidget(self.pickup_spoiler_pickup_combobox, 2, 2, 1, 1)

        self.pickup_spoiler_label = QLabel(self.pickup_tab)
        self.pickup_spoiler_label.setObjectName(u"pickup_spoiler_label")

        self.spoiler_pickup_layout.addWidget(self.pickup_spoiler_label, 2, 0, 1, 1)

        self.pickup_spoiler_show_all_button = QPushButton(self.pickup_tab)
        self.pickup_spoiler_show_all_button.setObjectName(u"pickup_spoiler_show_all_button")

        self.spoiler_pickup_layout.addWidget(self.pickup_spoiler_show_all_button, 2, 1, 1, 1)

        self.pickup_spoiler_scroll_area = QScrollArea(self.pickup_tab)
        self.pickup_spoiler_scroll_area.setObjectName(u"pickup_spoiler_scroll_area")
        self.pickup_spoiler_scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.pickup_spoiler_scroll_area.setSizeAdjustPolicy(QAbstractScrollArea.AdjustToContents)
        self.pickup_spoiler_scroll_area.setWidgetResizable(True)
        self.pickup_spoiler_scroll_area.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.pickup_spoiler_scroll_contents = QWidget()
        self.pickup_spoiler_scroll_contents.setObjectName(u"pickup_spoiler_scroll_contents")
        self.pickup_spoiler_scroll_contents.setGeometry(QRect(0, 0, 600, 345))
        self.verticalLayout_3 = QVBoxLayout(self.pickup_spoiler_scroll_contents)
        self.verticalLayout_3.setSpacing(6)
        self.verticalLayout_3.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(3, 0, 3, -1)
        self.pickup_spoiler_scroll_content_layout = QVBoxLayout()
        self.pickup_spoiler_scroll_content_layout.setSpacing(6)
        self.pickup_spoiler_scroll_content_layout.setObjectName(u"pickup_spoiler_scroll_content_layout")

        self.verticalLayout_3.addLayout(self.pickup_spoiler_scroll_content_layout)

        self.pickup_spoiler_scroll_area.setWidget(self.pickup_spoiler_scroll_contents)

        self.spoiler_pickup_layout.addWidget(self.pickup_spoiler_scroll_area, 4, 0, 1, 3)

        self.spoiler_starting_location_label = QLabel(self.pickup_tab)
        self.spoiler_starting_location_label.setObjectName(u"spoiler_starting_location_label")

        self.spoiler_pickup_layout.addWidget(self.spoiler_starting_location_label, 0, 0, 1, 3)

        self.spoiler_starting_items_label = QLabel(self.pickup_tab)
        self.spoiler_starting_items_label.setObjectName(u"spoiler_starting_items_label")

        self.spoiler_pickup_layout.addWidget(self.spoiler_starting_items_label, 1, 0, 1, 3)

        self.layout_info_tab.addTab(self.pickup_tab, "")

        self.verticalLayout.addWidget(self.layout_info_tab)

        SeedDetailsWindow.setCentralWidget(self.centralWidget)
        self.menuBar = QMenuBar(SeedDetailsWindow)
        self.menuBar.setObjectName(u"menuBar")
        self.menuBar.setGeometry(QRect(0, 0, 624, 21))
        SeedDetailsWindow.setMenuBar(self.menuBar)

        self.retranslateUi(SeedDetailsWindow)

        self.layout_info_tab.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(SeedDetailsWindow)
    # setupUi

    def retranslateUi(self, SeedDetailsWindow):
        SeedDetailsWindow.setWindowTitle(QCoreApplication.translate("SeedDetailsWindow", u"Seed Details", None))
        self.customize_user_preferences_button.setText(QCoreApplication.translate("SeedDetailsWindow", u"Customize in-game settings", None))
        self.progress_box.setTitle(QCoreApplication.translate("SeedDetailsWindow", u"Progress", None))
        self.stop_background_process_button.setText(QCoreApplication.translate("SeedDetailsWindow", u"Stop", None))
        self.progress_label.setText("")
        self.layout_description_left_label.setText(QCoreApplication.translate("SeedDetailsWindow", u"<html><head/><body><p>This content should have been replaced by code.</p></body></html>", None))
        self.layout_description_right_label.setText(QCoreApplication.translate("SeedDetailsWindow", u"<html><head/><body><p>This content should have been replaced by code.</p></body></html>", None))
        self.export_iso_button.setText(QCoreApplication.translate("SeedDetailsWindow", u"Save ISO", None))
        self.export_log_button.setText(QCoreApplication.translate("SeedDetailsWindow", u"Save Spoiler", None))
        self.layout_title_label.setText(QCoreApplication.translate("SeedDetailsWindow", u"<html><head/><body><p>Seed Hash: ????<br/>Preset Name: ???</p></body></html>", None))
        self.tool_button.setText(QCoreApplication.translate("SeedDetailsWindow", u"...", None))
        self.permalink_edit.setText(QCoreApplication.translate("SeedDetailsWindow", u"<insert permalink here>", None))
        self.permalink_label.setText(QCoreApplication.translate("SeedDetailsWindow", u"Permalink:", None))
        self.layout_info_tab.setTabText(self.layout_info_tab.indexOf(self.details_tab), QCoreApplication.translate("SeedDetailsWindow", u"Summary", None))
        self.pickup_spoiler_pickup_combobox.setItemText(0, QCoreApplication.translate("SeedDetailsWindow", u"None", None))

#if QT_CONFIG(tooltip)
        self.pickup_spoiler_label.setToolTip(QCoreApplication.translate("SeedDetailsWindow", u"Enter text to the right to filter the list below", None))
#endif // QT_CONFIG(tooltip)
        self.pickup_spoiler_label.setText(QCoreApplication.translate("SeedDetailsWindow", u"Search Pickup", None))
        self.pickup_spoiler_show_all_button.setText(QCoreApplication.translate("SeedDetailsWindow", u"Show All", None))
        self.spoiler_starting_location_label.setText(QCoreApplication.translate("SeedDetailsWindow", u"Starting Location", None))
        self.spoiler_starting_items_label.setText(QCoreApplication.translate("SeedDetailsWindow", u"Starting Items", None))
        self.layout_info_tab.setTabText(self.layout_info_tab.indexOf(self.pickup_tab), QCoreApplication.translate("SeedDetailsWindow", u"Spoiler: Pickups", None))
    # retranslateUi




import dataclasses
import json
from pathlib import Path
from unittest.mock import MagicMock

import pytest

from randovania.game_description import default_database
from randovania.game_description.game_description import GameDescription
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.item.item_database import ItemDatabase
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.games.game import RandovaniaGame
from randovania.games.prime import default_data
from randovania.interface_common.preset_manager import PresetManager
from randovania.layout.echoes_configuration import EchoesConfiguration
from randovania.layout.preset import Preset


@pytest.fixture
def test_files_dir() -> Path:
    return Path(__file__).parent.joinpath("test_files")


@pytest.fixture
def echo_tool(request, test_files_dir) -> Path:
    if request.config.option.skip_echo_tool:
        pytest.skip()
    return test_files_dir.joinpath("echo_tool.py")


@pytest.fixture()
def simple_data(test_files_dir: Path) -> dict:
    with test_files_dir.joinpath("small_game_data.json").open("r") as small_game_data:
        return json.load(small_game_data)


@pytest.fixture()
def preset_manager(tmpdir) -> PresetManager:
    return PresetManager(Path(tmpdir.join("presets")))


@pytest.fixture()
def default_preset(preset_manager) -> Preset:
    return preset_manager.default_preset.get_preset()


@pytest.fixture()
def default_layout_configuration(preset_manager) -> EchoesConfiguration:
    return preset_manager.default_preset.get_preset().configuration


@pytest.fixture()
def echoes_resource_database() -> ResourceDatabase:
    return default_database.resource_database_for(RandovaniaGame.PRIME2)


@pytest.fixture()
def echoes_item_database() -> ItemDatabase:
    return default_database.item_database_for_game(RandovaniaGame.PRIME2)


@pytest.fixture()
def echoes_game_data() -> dict:
    return default_data.decode_default_prime2()


@pytest.fixture()
def echoes_game_description(echoes_game_data) -> GameDescription:
    from randovania.game_description import data_reader
    return data_reader.decode_data(echoes_game_data)


@pytest.fixture()
def corruption_game_data() -> dict:
    return default_data.decode_default_prime3()


@pytest.fixture()
def corruption_game_description(corruption_game_data) -> GameDescription:
    from randovania.game_description import data_reader
    return data_reader.decode_data(corruption_game_data)


@pytest.fixture()
def randomizer_data() -> dict:
    return default_data.decode_randomizer_data()


class DataclassTestLib:
    def mock_dataclass(self, obj) -> MagicMock:
        return MagicMock(spec=[field.name for field in dataclasses.fields(obj)])


@pytest.fixture()
def dataclass_test_lib() -> DataclassTestLib:
    return DataclassTestLib()


@pytest.fixture()
def empty_patches() -> GamePatches:
    return GamePatches(0, {}, {}, {}, {}, {}, {}, None, {}, None)


def pytest_addoption(parser):
    parser.addoption('--skip-generation-tests', action='store_true', dest="skip_generation_tests",
                     default=False, help="Skips running layout generation tests")
    parser.addoption('--skip-resolver-tests', action='store_true', dest="skip_resolver_tests",
                     default=False, help="Skips running validation tests")
    parser.addoption('--skip-gui-tests', action='store_true', dest="skip_gui_tests",
                     default=False, help="Skips running GUI tests")
    parser.addoption('--skip-echo-tool', action='store_true', dest="skip_echo_tool",
                     default=False, help="Skips running tests that uses the echo tool")


try:
    import pytestqt


    @pytest.fixture()
    def skip_qtbot(request, qtbot):
        if request.config.option.skip_gui_tests:
            pytest.skip()
        return qtbot

except ImportError:
    @pytest.fixture()
    def skip_qtbot(request):
        pytest.skip()


def pytest_configure(config):
    if config.option.skip_generation_tests:
        setattr(config.option, 'markexpr', 'not skip_generation_tests')

    if config.option.skip_resolver_tests:
        setattr(config.option, 'markexpr', 'not skip_resolver_tests')

    if config.option.skip_gui_tests:
        setattr(config.option, 'markexpr', 'not skip_gui_tests')



from setuptools import setup, find_packages
from setuptools.command.build_py import build_py

try:
    from pyqt_distutils.build_ui import build_ui
except ModuleNotFoundError:
    build_ui = None


class custom_build_py(build_py):
    def run(self):
        self.run_command('build_ui')
        super().run()


with open("README.md") as readme_file:
    long_description = readme_file.read()


def version_scheme(version):
    import setuptools_scm.version
    if version.exact:
        return setuptools_scm.version.guess_next_simple_semver(
            version.tag, retain=setuptools_scm.version.SEMVER_LEN, increment=False)
    else:
        return version.format_next_version(
            setuptools_scm.version.guess_next_simple_semver, retain=setuptools_scm.version.SEMVER_MINOR
        )


setup(
    name='randovania',
    use_scm_version={
        "version_scheme": version_scheme,
        "local_scheme": "no-local-version",
        "write_to": "randovania/version.py",
    },
    author='Henrique Gemignani',
    url='https://github.com/randovania/randovania',
    description='A randomizer for the Metroid Prime 2: Echoes.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    cmdclass={
        "build_ui": build_ui,
        "build_py": custom_build_py
    },
    scripts=[
    ],
    package_data={
        "randovania": [
            "data/*",
            "data/ClarisPrimeRandomizer/*",
        ]
    },
    license='License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Games/Entertainment',
    ],
    python_requires=">=3.7",
    setup_requires=[
        "setuptools_scm>=3.5.0",
        "pyqt-distutils",
    ],
    install_requires=[
        'networkx',
        'bitstruct',
        'construct',
        'tenacity',
        'python-slugify',
        'python-socketio[asyncio_client]',
        'aiohttp[speedups]',
        'aiofiles',
    ],
    extras_require={
        "gui": [
            'PySide2>=5.12,<5.15',
            'appdirs',
            'asyncqt',
            'async-wiiload',
            'dolphin-memory-engine>=1.0.2',
            'markdown',
            'matplotlib>=3.3.3',
            'nod>=1.1',
            'pid>=3.0.0',
            'pypresence',
            'qdarkstyle',
        ],
        "server": [
            "cryptography",
            "discord.py",
            "eventlet",
            "flask-discord",
            "flask-socketio",
            "prometheus-flask-exporter",
            "peewee",
            "requests-oauthlib",
        ],
        "test": [
            'pytest',
            'pytest-cov',
            'pytest-qt',
            'pytest-asyncio',
            'pytest-mock',
            'mock>=4.0',
        ]
    },
    entry_points={
        'console_scripts': [
            "randovania = randovania.__main__:main"
        ]
    },
)
